#! /usr/bin/env python

"""IBIS-AMI model source code, AMI file, and IBIS file configuration utility.

Original author: David Banas

Original date:   February 26, 2016

Copyright (c) 2016 David Banas; all rights reserved World wide.

**Note:** The following use model has been deprecated!
The preferred approach is to make *executable* model configurators,
which draw what they need from this module.

This script gets called from a makefile, when any of the following need rebuilding:

* a C++ source code file
* a ``*.AMI`` file
* a ``*.IBS`` file
* a ``*.TST`` file (a dummy place-holder indicating that the test run config. files have been made)

All files will be rebuilt.
(We rebuild all files, because it doesn't take very long, and we can
ensure consistency this way.)

This gets triggered by one of two things:

#. The common model configuration information has changed, or
#. One of the EmPy template files was updated.

The idea here is that the ``*.IBS`` file, the ``*.AMI`` file, the C++ source file,
and the test run configuration files should be configured from a common model
configuration file, so as to ensure consistency between them all.
"""

import importlib.util
from datetime import date
from pathlib import Path
from typing import Any, NewType

import click
import em

ParamDict      = NewType("ParamDict",      dict[str, Any])
NamedParamDict = NewType("NamedParamDict", tuple[str, ParamDict])
TestDefinition = NewType("TestDefinition", tuple[str, NamedParamDict, NamedParamDict, str])

param_types = {
    "INT": {"c_type": "int", "ami_type": "Integer", "getter": "get_param_int"},
    "FLOAT": {"c_type": "double", "ami_type": "Float", "getter": "get_param_float"},
    "TAP": {"c_type": "double", "ami_type": "Tap", "getter": "get_param_float"},
    "BOOL": {"c_type": "bool", "ami_type": "Boolean", "getter": "get_param_bool"},
    "STRING": {"c_type": "char *", "ami_type": "String", "getter": "get_param_str"},
}


def print_param(indent, name, param):  # pylint: disable=too-many-branches
    """Print AMI parameter specification. Handle nested parameters, via
    recursion.

    Args:
        indent (str): String containing some number of spaces.
        name (str): Parameter name.
        param (dict): Dictionary containing parameter definition fields.
    """

    print(indent, f"({name}")
    if "subs" in param:
        for key in param["subs"]:
            print_param(indent + "    ", key, param["subs"][key])
        if "description" in param:
            print(indent + "    ", f"(Description {param['description']})")
    else:
        for fld_name, fld_key in [
            ("Usage", "usage"),
            ("Type", "type"),
            ("Format", "format"),
            ("Default", "default"),
            ("Description", "description"),
        ]:
            # Trap the special cases.
            if fld_name == "Type":
                print(indent, "    (Type", param_types[param["type"]]["ami_type"], ")")
            elif fld_name == "Default":
                if param["format"] == "Value":
                    pass
            elif fld_name == "Format":
                if param["format"] == "Value":
                    print(indent, "    (Value", param["default"], ")")
                elif param["format"] == "List":
                    print(indent, "    (List", end=" ")
                    for item in param["values"]:
                        print(item, end=" ")
                    print(")")
                    print(indent, "    (List_Tip", end=" ")
                    for item in param["labels"]:
                        print(item, end=" ")
                    print(")")
                else:
                    print(indent, f"    ({param['format']}", param["default"], param["min"], param["max"], ")")
            # Execute the default action.
            else:
                print(indent, f"    ({fld_name}", param[fld_key], ")")
    print(indent, ")")


def print_code(pname, param):
    """Print C++ code needed to query AMI parameter tree for a particular leaf.

    Args:
        pname (str): Parameter name.
        param (dict): Dictionary containing parameter definition fields.
    """

    print("       ", f'node_names.push_back("{pname}");')
    if "subs" in param:
        for key in param["subs"]:
            print_code(key, param["subs"][key])
    else:
        if param["usage"] == "In" or param["usage"] == "InOut":
            ptype = param["type"]
            print(f"        {param_types[ptype]['c_type']} {pname};")
            if ptype == "BOOL":
                print(f"        {pname} = {param_types[ptype]['getter']}(node_names, {param['default'].lower()});")
            else:
                print(f"        {pname} = {param_types[ptype]['getter']}(node_names, {param['default']});")
    print("       ", "node_names.pop_back();")


def mk_model(
    ibis_params: ParamDict,
    ami_params: ParamDict,
    model_name: str,
    description: str,
    out_dir: str = "."
) -> None:
    """
    Generate ibis, ami, and cpp files, by merging the
    device specific parameterization with the templates.

    Args:
        ibis_params: Dictionary of IBIS model parameter definitions.
        ami_params: Dictionary of AMI parameter definitions.
        model_name: Name given to IBIS model.
        description: Model description.

    Keyword Args:
        out_dir: Directory in which to place created files.
            Default: "."
    """

    py_file = (Path(out_dir).resolve() / model_name).with_suffix(".py")
    # Configure the model files.
    for ext in ["cpp", "ami", "ibs"]:
        out_file = py_file.with_suffix(f".{ext}")
        if ext == "ami":
            em_file = Path(__file__).parent.joinpath("generic.ami.em")
        elif ext == "ibs":
            em_file = Path(__file__).parent.joinpath("generic.ibs.em")
        else:
            em_file = out_file.with_suffix(".cpp.em")

        print(f"Building '{out_file}' from '{em_file}'...")
        with open(out_file, "w", encoding="utf-8") as o_file:
            interpreter = em.Interpreter(
                output=o_file,
                globals={
                    "ami_params": ami_params,
                    "ibis_params": ibis_params,
                    "param_types": param_types,
                    "model_name": model_name,
                    "description": description,
                    "date": str(date.today()),
                },
            )
            try:
                with open(em_file, "rt", encoding="utf-8") as in_file:
                    interpreter.file(in_file)
            finally:
                interpreter.shutdown()


def ami_config(py_file):
    """
    Read in ``py_file`` and cpp.em files, then generate: ibis, ami, and cpp files.

    Args:
        py_file: name of model configuration file (<stem>.py)

    Notes:
        1. This function is deprecated! Instead, make your model configurator executable
        and import what you need from this module. This is much cleaner.
    """

    file_base_name = Path(py_file).stem

    # Read model configuration information.
    print(f"Reading model configuration information from file: {py_file}.")
    spec = importlib.util.spec_from_file_location(file_base_name, py_file)
    cfg = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(cfg)

    mk_model(cfg.ibis_params, cfg.ami_params, cfg.kFileBaseName, cfg.kDescription, out_dir=Path(py_file).parent)


def mk_combs(dict_items: list[tuple[str, Any]]) -> list[list[tuple[str, Any]]]:
    """
    Make all combinations possible from a list of dictionary items.

    Args:
        dict_items: List of dictionary key/value pairs.
            The values are lists.

    Return:
        List of all possible combinations of key values.
    """
    if not dict_items:
        return [
            [],
        ]
    head, *tail = dict_items
    k, vs = head
    kvals = [(k, v) for v in vs]
    return [[kval] + rest for kval in kvals for rest in mk_combs(tail)]


def mk_tests(  # pylint: disable=too-many-locals
    test_defs: dict[str, TestDefinition],
    file_base_name: str,
    test_dir: str = "test_runs"
) -> None:
    """
    Make the test run configuration files.

    Args:
        test_defs: Dictionary of test sweep definitions.
        file_base_name: Stem name for test run definition files.

    Keyword Args:
        test_dir: Directory in which to place the created test run definition files.
            Default: "test_runs/"
    """

    pname = Path(test_dir).resolve()
    pname.mkdir(exist_ok=True)
    pname = (pname / file_base_name).resolve()
    pname.mkdir(exist_ok=True)
    for fname in test_defs.keys():
        desc, ami_defs, sim_defs, ref_fstr = test_defs[fname]
        ami_str, ami_dict = ami_defs
        sim_str, sim_dict = sim_defs
        with open((pname / fname).with_suffix(".run"), "w", encoding="utf-8") as f:
            f.write(desc + "\n")
            for ami_comb in mk_combs(list(ami_dict.items())):
                for sim_comb in mk_combs(list(sim_dict.items())):
                    pdict = dict(ami_comb)
                    pdict.update(dict(sim_comb))
                    f.write(f"\n('{ami_str.format(pdict=pdict)}_{sim_str.format(pdict=pdict)}', \\\n")
                    f.write(f"  ({{'root_name': '{file_base_name}', \\\n")
                    for k, v in ami_comb:
                        f.write(f"    '{k}': {v}, \\\n")
                    f.write("   }, \\\n")
                    if sim_comb:
                        head, *tail = sim_comb
                        k, v = head
                        f.write(f"   {{'{k}': {v}, \\\n")
                        for k, v in tail:
                            f.write(f"    '{k}': {v}, \\\n")
                        f.write("   } \\\n")
                    f.write("  ), \\\n")
                    if ref_fstr:
                        f.write(f"  '{ref_fstr.format(pdict=pdict)}', \\\n")
                    f.write(")\n")


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("py_file", type=click.Path(exists=True, resolve_path=True))
@click.version_option()
def main(py_file):
    """
    Configure IBIS-AMI model C++ source code, IBIS model, and AMI file.
    This command generates three files based off the input config file.
    It expects a .cpp.em file to be located in the same directory so that it can
    generate a cpp file from the config file and template file.

    Args:
       py_file: name of model configuration file (*.py)

    Notes:
        1. This command is deprecated! Instead, make your model configurator executable
        and import what you need from this module. This is much cleaner.
    """
    ami_config(py_file)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
Definitions common to all PyIBIS-AMI modules.

Original author: David Banas <capn.freako@gmail.com>

Original date:   May 15, 2024

Copyright (c) 2024 David Banas; all rights reserved World wide.
"""

from typing             import Any, TypeAlias, TypeVar
import numpy.typing as npt  # type: ignore
from scipy.linalg       import convolution_matrix, lstsq

Real = TypeVar("Real", float, float)
Comp = TypeVar("Comp", complex, complex)
Rvec: TypeAlias = npt.NDArray["Real"]
Cvec: TypeAlias = npt.NDArray["Comp"]

PI:    float = 3.141592653589793238462643383279502884
TWOPI: float = 2.0 * PI

# TestConfig: TypeAlias = tuple[str, tuple[dict[str, Any], dict[str, Any]]]
# TestSweep:  TypeAlias = tuple[str, str, list[TestConfig]]
TestConfig = tuple[str, tuple[dict[str, Any], dict[str, Any]]]
TestSweep = tuple[str, str, list[TestConfig]]


def deconv_same(y: Rvec, x: Rvec) -> Rvec:
    """
    Deconvolve input from output, to recover filter response, for same length I/O.

    Args:
        y: output signal
        x: input signal

    Returns:
        h: filter impulse response.
    """
    A = convolution_matrix(x, len(y), "same")
    h, _, _, _ = lstsq(A, y)
    return h



"""A package of Python modules, used to configure and test IBIS-AMI models."""

print("pyibisami is a package of sub-modules, and is not directly executable.")
print("Perhaps, you meant:")
print("    - pyibisami.ami.config")
print("    - pyibisami.tools.run_tests")
raise RuntimeError("pyibisami is not executable.")



"""
IBIS-AMI reserved parameter names, as a Python ``dataclass``.

Original author: David Banas <capn.freako@gmail.com>

Original date:   March 17, 2025

Copyright (c) 2025 David Banas; all rights reserved World wide.
"""

from dataclasses import dataclass


@dataclass
class AmiReservedParameterName():
    "IBIS-AMI Reserved Parameter Name"

    pname: str

    def __post_init__(self):
        "Validate parameter name."

        if self.pname not in RESERVED_PARAM_NAMES:
            raise ValueError(f"Parameter name: {self.pname}, is not an IBIS-AMI reserved parameter name!")


RESERVED_PARAM_NAMES = [
    "AMI_Version",
    "Init_Returns_Impulse",
    "GetWave_Exists",
    "Use_Init_Output",
    "Max_Init_Aggressors",
    "Ignore_Bits",
    "Resolve_Exists",
    "Model_Name",
    "Special_Param_Names",
    "Component_Name",
    "Signal_Name",
    "Rx_Decision_Time",
    "DC_Offset",
    "Rx_Use_Clock_Input",
    "Supporting_Files",
    "DLL_Path",
    "DLL_ID",
    "Tx_Jitter",
    "Tx_DCD",
    "Tx_Rj",
    "Tx_Dj",
    "Tx_Sj",
    "Tx_Sj_Frequency",
    "Rx_DCD",
    "Rx_Rj",
    "Rx_Dj",
    "Rx_Sj",
    "Rx_Clock_PDF",
    "Rx_Clock_Recovery_Mean",
    "Rx_Clock_Recovery_Rj",
    "Rx_Clock_Recovery_Dj",
    "Rx_Clock_Recovery_Sj",
    "Rx_Clock_Recovery_DCD",
    "Rx_Receiver_Sensitivity",
    "Rx_Noise",
    "Rx_GaussianNoise",
    "Rx_UniformNoise",
    "Modulation",
    "PAM4_Mapping",
    "PAM4_UpperThreshold",
    "PAM4_CenterThreshold",
    "PAM4_LowerThreshold",
    "PAM4_UpperEyeOffset",
    "PAM4_CenterEyeOffset",
    "PAM4_LowerEyeOffset",
    "Repeater_Type",
    "BCI_Protocol",
    "BCI_ID",
    "BCI_State",
    "BCI_Message_Interval_UI",
    "BCI_Training_UI",
    "BCI_Training_Mode",
    "Ts4file",
    "Tx_V",
    "Tx_R",
    "Rx_R",
]



"""``AMIParameter`` class definition, plus some helpers.

Original author: David Banas <capn.freako@gmail.com>

Original date:   December 24, 2016

Copyright (c) 2019 David Banas; all rights reserved World wide.
"""

#####
# AMI parameter
#####


class AMIParamError(Exception):
    """Base Exception for all AMI Parameter Errors."""


class AMIParameter:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """IBIS-AMI model parameter.

    This class encapsulates the attributes and behavior of a AMI
    parameter.
    """

    # Properties.

    # Note: They are read-only, despite the presence of apparent setters.
    #       (The idea is that, once initialized, parameter definitions
    #        are immutable.)
    #       The only exception to this is pvalue, which has been made
    #       writable, for scripted non-GUI use cases.
    #       Be very careful w/ this; there is NO CHECKING!

    # Note that the setters, below, are only intended for use by
    # __init__(). They may raise an *AMIParamError* exception. This is
    # to ensure that a malformed instance never be created.

    def _get_name(self):
        """pname."""
        return self._name

    pname = property(_get_name, doc="Name of AMI parameter.")

    # pusage
    def _set_usage(self, values):
        "Process *Usage* tag."

        val = values[0]
        if val in ("In", "Out", "InOut", "Info"):
            self._usage = val
        else:
            raise AMIParamError(f"Unrecognized usage value: '{val}'.")

    def _get_usage(self):
        return self._usage

    pusage = property(_get_usage, doc="Value of AMI parameter 'Usage' tag.")

    # ptype
    def _set_type(self, values):
        "Process *Type* tag."

        val = values[0]
        if val in ("Float", "Integer", "String", "Boolean", "UI", "Tap"):
            self._type = val
        else:
            raise AMIParamError(f"Unrecognized type value: '{val}'.")

    def _get_type(self):
        return self._type

    ptype = property(_get_type, doc="Value of AMI parameter 'Type' tag.")

    # pformat
    def _set_format(self, values):
        "Process *Format* tag."

        form = values[0]
        if len(values) < 2:
            raise AMIParamError(f"No values provided for: '{form}'.")
        self._format = form
        self._format_rem = values[1:]

    def _get_format(self):
        return self._format

    pformat = property(_get_format, doc="Value of AMI parameter 'Format' tag.")

    # pvalue
    def _get_value(self):
        return self._value

    def _set_val(self, new_val):
        self._value = new_val

    pvalue = property(_get_value, _set_val, doc="Value of AMI parameter.")

    # pmin
    def _get_min(self):
        return self._min

    pmin = property(_get_min, doc="Minimum value of AMI parameter.")

    # pmax
    def _get_max(self):
        return self._max

    pmax = property(_get_max, doc="Maximum value of AMI parameter.")

    # default
    def _set_default(self, values):
        """Process *Default* tag."""
        self._default = values[0]

    def _get_default(self):
        return self._default

    pdefault = property(_get_default, doc="Default value of AMI parameter.")

    # pdescription
    def _set_description(self, values):
        """Process *Description* tag."""
        self._description = values[0]

    def _get_description(self):
        return self._description

    pdescription = property(_get_description, doc="Description of AMI parameter.")

    # plist_tip
    def _set_list_tip(self, values):
        """Process *List_Tip* tag."""
        self._list_tip = [x.strip('"') for x in values]

    def _get_list_tip(self):
        return self._list_tip

    plist_tip = property(_get_list_tip, doc="List tips of AMI parameter.")

    # Helpers.
    # These 3 just accomodate the awkwardness caused by the optional
    # nature of the 'Format' keyword, in *.AMI files. Since, a properly
    # formed instance of *AMIParameter* will always have a *format*
    # property, we don't need to make these properties of that class.

    # value
    def _set_value(self, values):
        """Process *Value* tag."""

        return self._set_format(["Value"] + values)

    # range
    def _set_range(self, values):
        """Process *Range* tag."""

        return self._set_format(["Range"] + values)

    # usage
    def _set_list(self, values):
        """Process *List* tag."""

        return self._set_format(["List"] + values)

    # Holds any warnings encountered, during initialization.
    # (Any errors encountered will prevent initialization from completing.)
    _msg = ""

    def _get_msg(self):
        return self._msg

    msg = property(_get_msg, doc="Any warning messages encountered, during parameter initialization.")

    # This dictionary defines both:
    #
    #  - the allowed parameter definition tag names, and
    #  - their processing functions.
    #
    # The idea is to allow this class to grow along with the IBIS
    # standard, without having to change any of its boilerplate.

    _param_def_tag_procs = {
        "Usage": _set_usage,
        "Type": _set_type,
        "Format": _set_format,
        "Value": _set_value,
        "Range": _set_range,
        "List": _set_list,
        "Corner": _set_list,
        "Default": _set_default,
        "Description": _set_description,
        "List_Tip": _set_list_tip,
        "Label": _set_list_tip,
        "Labels": _set_list_tip,
    }

    def __init__(self, name, tags):  # pylint: disable=too-many-branches,too-many-statements
        """
        Args:
            name (str): The name of the AMI parameter being created.
            tags ([(str, [a])]): A list of pairs, each containing

                - a parameter definition tag name
                    (Must be one of the keys from the '_param_def_tag_procs' dictionary.)

                - a list of values to be associated with that tag.
        """
        # Initialization
        self._usage = None
        self._type = None
        self._format = None
        self._value = None
        self._min = None
        self._max = None
        self._default = None
        self._description = ""
        self._list_tip = None

        # Holds any warnings encountered, during initialization.
        # (Any errors encountered will prevent initialization from completing.)
        self._msg = ""

        # Process all parameter definition tags.
        for tag in tags:
            tag_name = tag[0]
            if tag_name in self._param_def_tag_procs:
                try:
                    self._param_def_tag_procs[tag_name](self, tag[1])
                except AMIParamError as err:
                    raise AMIParamError(f"Problem initializing parameter, '{name}': {err}\n") from err

        # Validate and complete the instance.
        # Check for required tags.
        param_usage = self._usage
        param_type = self._type
        param_format = self._format
        param_default = self._default
        if param_usage is None:
            raise AMIParamError("Missing 'Usage' tag!\n")
        if param_type is None:
            raise AMIParamError("Missing 'Type' tag!\n")
        if param_format is None:
            if param_default is None:
                raise AMIParamError("Missing both 'Format' and 'Default' tags!\n")
            self._value = param_default
            param_format = "Value"
            self._format_rem = [param_default]
        # Check for mutual exclusivity of 'Format Value' and 'Default'.
        elif (param_format == "Value") and (param_default is not None):
            self._msg += "'Format Value' and 'Default' both found! (They are mutually exclusive.)\n"
        # Canonicalize Boolean if necessary.
        elif (param_format == "List") and (param_type == "Boolean"):
            if param_default:
                self._value = param_default
            else:
                self._value = "False"
            param_format = "Value"
            self._format_rem = [self._value]
        self._format = param_format

        # Check for 'Default' used with parameter type 'Out'.
        if (param_usage == "Out") and (param_default is not None):
            raise AMIParamError("'Default' may not be used with parameter type 'Out'!\n")

        # Complete the instance.
        vals = self._format_rem
        if param_format == "Value":
            value_str = vals[0].strip()
            if param_type in ("Float", "UI"):
                try:
                    self._value = float(value_str)
                except (ValueError, TypeError) as exc:
                    raise AMIParamError(f"Couldn't read float from '{value_str}'.\n") from exc
            elif param_type == "Integer":
                try:
                    self._value = int(float(value_str))  # Hack to accommodate: "1e5", for instance.
                except (ValueError, TypeError) as exc:
                    raise AMIParamError(f"Couldn't read integer from '{value_str}'.\n") from exc
            elif param_type == "Boolean":
                if value_str == "True":
                    self._value = True
                elif value_str == "False":
                    self._value = False
                else:
                    raise AMIParamError(f"Couldn't read Boolean from '{value_str}'.\n")
            else:  # `String`
                self._value = value_str.strip('"')
        elif param_format == "Range":
            if param_type not in ("Float", "Integer", "UI", "Tap"):
                raise AMIParamError(f"Illegal type, '{param_type}', for use with Range.\n")
            if len(vals) < 3:
                raise AMIParamError(f"Insufficient number of values, {len(vals)}, provided for Range.\n")
            if param_type in ("Float", "UI", "Tap"):
                try:
                    temp_vals = list(map(float, vals[:3]))
                except (ValueError, TypeError) as exc:
                    raise AMIParamError(f"Couldn't read floats from '{vals[:3]}'.\n") from exc
            else:
                try:
                    temp_vals = list(map(int, vals[:3]))
                except (ValueError, TypeError) as exc:
                    raise AMIParamError(f"Couldn't read integers from '{vals[:3]}'.\n") from exc
            self._value = temp_vals[0]
            self._min = temp_vals[1]
            self._max = temp_vals[2]
        else:  # param_format == 'List'
            if param_type in ("Float", "UI"):
                try:
                    temp_vals = list(map(float, vals))
                except (ValueError, TypeError) as exc:
                    raise AMIParamError(f"Couldn't read floats from '{vals}'.\n") from exc
            elif param_type in ("Integer", "Tap"):
                try:
                    temp_vals = list(map(int, vals))
                except (ValueError, TypeError) as exc:
                    raise AMIParamError(f"Couldn't read integers from '{vals}'.\n") from exc
            else:  # 'param_type' == 'String'
                try:
                    temp_vals = list(map(str, vals))
                    temp_vals = [x.strip('"') for x in temp_vals]
                except (ValueError, TypeError) as exc:
                    raise AMIParamError(f"Couldn't read strings from '{vals}'.\n") from exc
            self._value = temp_vals

        self._name = name



"""IBIS-AMI parameter parsing and configuration utilities.

Original author: David Banas <capn.freako@gmail.com>

Original date:   December 17, 2016

Copyright (c) 2019 David Banas; all rights reserved World wide.
"""

from ctypes         import c_double
import re
from typing         import Any, Callable, NewType, Optional, TypeAlias

import numpy as np
from numpy.typing import NDArray
from parsec import ParseError, generate, many, regex, string
from traits.api import Bool, Enum, HasTraits, Range, Trait, TraitType
from traitsui.api import Group, HGroup, Item, VGroup, View
from traitsui.menu import ModalButtons

from .model                     import AMIModelInitializer
from .parameter                 import AMIParamError, AMIParameter
from .reserved_parameter_names  import AmiReservedParameterName, RESERVED_PARAM_NAMES

# New types and aliases.
# Parameters  = NewType('Parameters',  dict[str, AMIParameter] | dict[str, 'Parameters'])
# ParamValues = NewType('ParamValues', dict[str, list[Any]]    | dict[str, 'ParamValues'])
# See: https://stackoverflow.com/questions/70894567/using-mypy-newtype-with-type-aliases-or-protocols
ParamName  = NewType("ParamName", str)
ParamValue:  TypeAlias = int | float | str | list["ParamValue"]
Parameters:  TypeAlias = dict[ParamName, "'AMIParameter' | 'Parameters'"]
ParamValues: TypeAlias = dict[ParamName, "'ParamValue'   | 'ParamValues'"]

AmiName = NewType("AmiName", str)
AmiAtom: TypeAlias = bool | int | float | str
AmiExpr: TypeAlias = "'AmiAtom' | 'AmiNode'"
AmiNode: TypeAlias = tuple[AmiName, list[AmiExpr]]
AmiNodeParser: TypeAlias = Callable[[str], AmiNode]
AmiParser:     TypeAlias = Callable[[str], tuple[AmiName, list[AmiNode]]]  # Atoms may not exist at the root level.

ParseErrMsg = NewType("ParseErrMsg", str)
AmiRootName = NewType("AmiRootName", str)
ReservedParamDict: TypeAlias = dict[AmiReservedParameterName, AMIParameter]
ModelSpecificDict: TypeAlias = dict[ParamName, "'AMIParameter' | 'ModelSpecificDict'"]

__all__ = [
    "ParamName", "ParamValue", "Parameters", "ParamValues",
    "AmiName", "AmiAtom", "AmiExpr", "AmiNode", "AmiNodeParser", "AmiParser",
    "ami_parse", "AMIParamConfigurator"]

#####
# AMI parameter configurator.
#####


class AMIParamConfigurator(HasTraits):
    """Customizable IBIS-AMI model parameter configurator.

    This class can be configured to present a customized GUI to the user
    for configuring a particular IBIS-AMI model.

    The intended use model is as follows:

     1. Instantiate this class only once per IBIS-AMI model invocation.
        When instantiating, provide the unprocessed contents of the AMI
        file, as a single string. This class will take care of getting
        that string parsed properly, and report any errors or warnings
        it encounters, in its ``ami_parsing_errors`` property.

     2. When you want to let the user change the AMI parameter
        configuration, call the ``open_gui`` member function.
        (Or, just call the instance as if it were a function.)
        The instance will then present a GUI to the user,
        allowing him to modify the values of any *In* or *InOut* parameters.
        The resultant AMI parameter dictionary, suitable for passing
        into the ``ami_params`` parameter of the ``AMIModelInitializer``
        constructor, can be accessed, via the instance's
        ``input_ami_params`` property. The latest user selections will be
        remembered, as long as the instance remains in scope.

    The entire AMI parameter definition dictionary, which should *not* be
    passed to the ``AMIModelInitializer`` constructor, is available in the
    instance's ``ami_param_defs`` property.

    Any errors or warnings encountered while parsing are available, in
    the ``ami_parsing_errors`` property.
    """

    def __init__(self, ami_file_contents_str: str) -> None:
        """
        Args:
            ami_file_contents_str: The unprocessed contents of the AMI file, as a single string.
        """

        # Super-class initialization is ABSOLUTELY NECESSARY, in order
        # to get all the Traits/UI machinery setup correctly.
        super().__init__()

        # Parse the AMI file contents, storing any errors or warnings, and customize the view accordingly.
        (err_str,
         root_name,
         description,
         reserved_param_dict,
         model_specific_dict) = parse_ami_file_contents(ami_file_contents_str)
        if err_str:
            raise RuntimeError(
                "\n".join([
                    "AMI parsing error:",
                    err_str
                ]))
        if not reserved_param_dict:
            raise ValueError(
                "\n".join([
                    "No 'Reserved_Parameters' section found!",
                    err_str
                ]))
        if not model_specific_dict:
            raise ValueError(
                "\n".join([
                    "No 'Model_Specific' section found!",
                    err_str
                ]))
        gui_items, new_traits = make_gui(model_specific_dict)
        trait_names = []
        for trait in new_traits:
            self.add_trait(trait[0], trait[1])
            trait_names.append(trait[0])
        self._param_trait_names = trait_names
        self._root_name = root_name
        self._ami_parsing_errors = err_str
        self._content = gui_items
        self._reserved_param_dict = reserved_param_dict
        self._model_specific_dict = model_specific_dict
        self._description = description

    def __call__(self):
        self.open_gui()

    def open_gui(self):
        """Present a customized GUI to the user, for parameter
        customization."""
        # self.configure_traits(kind='modal')  # Waiting for Enthought/Traits PR1841 to be accepted.
        self.configure_traits()

    def default_traits_view(self):
        "Default Traits/UI view definition."
        view = View(
            resizable=False,
            buttons=ModalButtons,
            title=f"{self._root_name} AMI Parameter Configurator",
        )
        view.set_content(self._content)
        return view

    def fetch_param(self, branch_names):
        """Returns the parameter found by traversing 'branch_names' or None if
        not found.

        Note: 'branch_names' should *not* begin with 'root_name'.
        """
        param_dict = self.ami_param_defs
        while branch_names:
            branch_name = branch_names.pop(0)
            if branch_name in param_dict:
                param_dict = param_dict[branch_name]
            else:
                return None
        if isinstance(param_dict, AMIParameter):
            return param_dict
        return None

    def fetch_param_val(self, branch_names):
        """Returns the value of the parameter found by traversing
        'branch_names' or None if not found.

        Note: 'branch_names' should *not* begin with 'root_name'.
        """
        _param = self.fetch_param(branch_names)
        if _param:
            return _param.pvalue
        return None

    def set_param_val(self, branch_names, new_val):
        """Sets the value of the parameter found by traversing 'branch_names'
        or raises an exception if not found.

        Note: 'branch_names' should *not* begin with 'root_name'.
        Note: Be careful! There is no checking done here!
        """

        param_dict = self.ami_param_defs
        while branch_names:
            branch_name = branch_names.pop(0)
            if branch_name in param_dict:
                param_dict = param_dict[branch_name]
            else:
                raise ValueError(
                    f"Failed parameter tree search looking for: {branch_name}; available keys: {param_dict.keys()}"
                )
        if isinstance(param_dict, AMIParameter):
            param_dict.pvalue = new_val
            try:
                eval(f"self.set({branch_name}_={new_val})")  # pylint: disable=eval-used
            except Exception:  # pylint: disable=broad-exception-caught
                eval(f"self.set({branch_name}={new_val})")  # pylint: disable=eval-used
        else:
            raise TypeError(f"{param_dict} is not of type: AMIParameter!")

    @property
    def ami_parsing_errors(self):
        """Any errors or warnings encountered, while parsing the AMI parameter
        definition file contents."""
        return self._ami_parsing_errors

    @property
    def ami_param_defs(self) -> dict[str, ReservedParamDict | ModelSpecificDict]:
        """The entire AMI parameter definition dictionary.

        Should *not* be passed to ``AMIModelInitializer`` constructor!
        """
        return {"Reserved_Parameters": self._reserved_param_dict,
                "Model_Specific": self._model_specific_dict}

    @property
    def input_ami_params(self) -> ParamValues:
        """
        The dictionary of *Model Specific* AMI parameters of type 'In' or
        'InOut', along with their user selected values.

        Should be passed to ``AMIModelInitializer`` constructor.
        """

        res: ParamValues = {}
        res[ParamName("root_name")] = str(self._root_name)
        params = self._model_specific_dict
        for pname in params:
            res.update(self.input_ami_param(params, pname))
        return res

    def input_ami_param(
        self,
        params: Parameters,
        pname: ParamName,
        prefix: str = ""
    ) -> ParamValues:
        """
        Retrieve one AMI parameter value, or dictionary of subparameter values,
        from the given parameter definition dictionary.

        Args:
            params: The parameter definition dictionary.
            pname: The simple name of the parameter of interest, used by the IBIS-AMI model.

        Keyword Args:
            prefix: The current working parameter name prefix.

        Returns:
            A dictionary of parameter values indexed by non-prefixed parameter names.

        Notes:
            1. The "prefix" referred to above refers to a string encoding of the
            hierarchy above a particular trait. We need this hierarchy for the
            sake of the ``Traits/UI`` machinery, which addresses traits by name
            alone. However, the IBIS-AMI model is not expecting it. So, we have
            to strip it off, before sending the result here into ``AMI_Init()``.
        """

        res = {}
        tname = prefix + pname     # This is the fully hierarchical trait name, used by the Traits/UI machinery.
        param = params[pname]
        if isinstance(param, AMIParameter):
            if tname in self._param_trait_names:  # If model specific and of type In or InOut...
                # See the docs on the *HasTraits* class, if this is confusing.
                # Querry for a mapped trait, first, by trying to get '<trait_name>_'. (Note the underscore.)
                try:
                    res[pname] = self.trait_get(tname + "_")[tname + "_"]
                # If we get an exception, we have an ordinary (i.e. - not mapped) trait.
                except Exception:  # pylint: disable=broad-exception-caught
                    res[pname] = self.trait_get(tname)[tname]
        elif isinstance(param, dict):  # We received a dictionary of subparameters, in 'param'.
            subs: ParamValues = {}
            for sname in param:
                subs.update(self.input_ami_param(param, sname, prefix=pname + "_"))  # type: ignore
            res[pname] = subs
        return res

    @property
    def info_ami_params(self):
        "Dictionary of *Reserved* AMI parameter values."
        return self._reserved_param_dict

    def get_init(
        self,
        bit_time:         float,
        sample_interval:  float,
        channel_response: NDArray[np.longdouble],
        ami_params: Optional[dict[str, Any]] = None
    ) -> AMIModelInitializer:
        """
        Get a model initializer, configured by the user if necessary.
        """

        row_size = len(channel_response)
        if ami_params:
            initializer = AMIModelInitializer(
                ami_params,
                info_params=self.info_ami_params,
                bit_time=c_double(bit_time),
                row_size=row_size,
                sample_interval=c_double(sample_interval)
            )
        else:
            # This call will invoke a GUI applet for the user to interact with,
            # to configure the AMI parameter values.
            self()
            initializer = AMIModelInitializer(
                self.input_ami_params,
                info_params=self.info_ami_params,
                bit_time=c_double(bit_time),
                row_size=row_size,
                sample_interval=c_double(sample_interval)
            )

        # Don't try to pack this into the parentheses above!
        initializer.channel_response = channel_response
        return initializer


#####
# AMI file parser.
#####

# ignore cases.
whitespace = regex(r"\s+", re.MULTILINE)
comment = regex(r"\|.*")
ignore = many(whitespace | comment)


def lexeme(p):
    """Lexer for words."""
    return p << ignore  # skip all ignored characters.


def int2tap(x):
    """Convert integer to tap position."""
    x = x.strip()
    match x[0]:
        case "-":
            res = "pre" + x[1:]
        case "+":
            res = "post" + x[1:]
        case _:
            res = "post" + x
    return res


lparen = lexeme(string("("))
rparen = lexeme(string(")"))
number = lexeme(regex(r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?"))
integ = lexeme(regex(r"[-+]?[0-9]+"))
nat = lexeme(regex(r"[0-9]+"))
# tap_ix = (integ << whitespace).parsecmap(int2tap)  # Doesn't work!
# tap_ix = integ.parsecmap(int2tap)  # Breaks parsing of node names that begin with a numeral!
tap_ix = integ.parsecmap(int2tap) << whitespace
symbol = lexeme(regex(r"[0-9a-zA-Z_][^\s()]*"))
true = lexeme(string("True")).result(True)
false = lexeme(string("False")).result(False)
ami_string = lexeme(regex(r'"[^"]*"'))

atom = number | symbol | ami_string | (true | false)
node_name = tap_ix ^ symbol  # `tap_ix` is new and gives the tap position; negative positions are allowed.


@generate("AMI node")
def node():
    "Parse AMI node."
    yield lparen
    label  = yield node_name
    values = yield many(expr)
    yield rparen
    return (label, values)


@generate("AMI file")
def root():
    "Parse AMI file."
    yield lparen
    label  = yield node_name
    values = yield many(node)
    yield rparen
    return (label, values)


expr = atom | node
ami = ignore >> root
ami_parse: AmiParser = ami.parse


def proc_branch(branch):
    """Process a branch in a AMI parameter definition tree.

    That is, build a dictionary from a pair containing:
        - a parameter name, and
        - a list of either:
            - parameter definition tags, or
            - subparameters.

    We distinguish between the two possible kinds of payloads, by
    peaking at the names of the first two items in the list and noting
    whether they are keys of 'AMIParameter._param_def_tag_procs'.
    We have to do this twice, due to the dual use of the 'Description'
    tag and the fact that we have no guarantee of any particular
    ordering of subparameter branch items.

    Args:
        p (str, list): A pair, as described above.

    Returns:
        (str, dict): A pair containing:
            err_str: String containing any errors or warnings encountered,
                while building the parameter dictionary.
            param_dict: Resultant parameter dictionary.
    """
    results = ("", {})  # Empty Results
    if len(branch) != 2:
        if not branch:
            err_str = "ERROR: Empty branch provided to proc_branch()!\n"
        else:
            err_str = f"ERROR: Malformed item: {branch[0]}\n"
        results = (err_str, {})

    param_name = branch[0]
    param_tags = branch[1]

    if not param_tags:
        err_str = f"ERROR: No tags/subparameters provided for parameter, '{param_name}'\n"
        results = (err_str, {})

    try:
        if (
            (len(param_tags) > 1)
            and (  # noqa: W503
                param_tags[0][0] in AMIParameter._param_def_tag_procs  # pylint: disable=protected-access  # noqa: W503
            )
            and (  # noqa: W503
                param_tags[1][0] in AMIParameter._param_def_tag_procs  # pylint: disable=protected-access  # noqa: W503
            )
        ):
            try:
                results = ("", {param_name: AMIParameter(param_name, param_tags)})
            except AMIParamError as err:
                results = (str(err), {})
        elif param_name == "Description":
            results = ("", {"description": param_tags[0].strip('"')})
        else:
            err_str = ""
            param_dict = {}
            param_dict[param_name] = {}
            for param_tag in param_tags:
                temp_str, temp_dict = proc_branch(param_tag)
                param_dict[param_name].update(temp_dict)
                if temp_str:
                    err_str = (
                        f"Error returned by recursive call, while processing parameter, '{param_name}':\n{temp_str}"
                    )
                    results = (err_str, param_dict)

            results = (err_str, param_dict)
    except Exception:  # pylint: disable=broad-exception-caught
        print(f"Error processing branch:\n{param_tags}")
    return results


def parse_ami_file_contents(  # pylint: disable=too-many-locals,too-many-branches
    file_contents: str
) -> tuple[ParseErrMsg, AmiRootName, str, ReservedParamDict, ModelSpecificDict]:
    """
    Parse the contents of an IBIS-AMI *parameter definition* (i.e. - `*.ami`) file.

    Args:
        file_contents: The contents of the file, as a single string.

    Example:
        ::

            with open(<ami_file_name>) as ami_file:
                file_contents = ami_file.read()
                (err_str, root_name, reserved_param_dict, model_specific_param_dict) = parse_ami_file_contents(file_contents)

    Returns:
        A tuple containing

            1. Any error message generated by the parser. (empty on success)

            2. AMI file "root" name.

            3. *Reserved Parameters* dictionary. (empty on failure)

                - The keys of the *Reserved Parameters* dictionary are
                limited to those called out in the IBIS-AMI specification.

                - The values of the *Reserved Parameters* dictionary
                must be instances of class ``AMIParameter``.

            4. *Model Specific Parameters* dictionary. (empty on failure)

                - The keys of the *Model Specific Parameters* dictionary can be anything.

                - The values of the *Model Specific Parameters* dictionary
                may be either: an instance of class ``AMIParameter``, or a nested sub-dictionary.
    """
    try:
        res = ami_parse(file_contents)
    except ParseError as pe:
        err_str = ParseErrMsg(f"Expected {pe.expected} at {pe.loc()} in:{pe.text[pe.index: pe.index + 20]}")
        return err_str, AmiRootName(""), "", {}, {}

    err_str, param_dict = proc_branch(res)
    if err_str:
        return (err_str, AmiRootName(""), "", {}, {})
    if len(param_dict.keys()) != 1:
        raise ValueError(f"Malformed AMI parameter S-exp has top-level keys: {param_dict.keys()}!")

    reserved_found = False
    init_returns_impulse_found = False
    getwave_exists_found = False
    model_spec_found = False
    root_name, params = list(param_dict.items())[0]
    description = ""
    reserved_params_dict = {}
    model_specific_dict = {}
    _err_str = ""
    for label in list(params.keys()):
        tmp_params = params[label]
        if label == "Reserved_Parameters":
            reserved_found = True
            for param_name in list(tmp_params.keys()):
                if param_name not in RESERVED_PARAM_NAMES:
                    _err_str += f"WARNING: Unrecognized reserved parameter name, '{param_name}', found in parameter definition string!\n"
                    continue
                param = tmp_params[param_name]
                if param.pname == "AMI_Version":
                    if param.pusage != "Info" or param.ptype != "String":
                        _err_str += "WARNING: Malformed 'AMI_Version' parameter.\n"
                elif param.pname == "Init_Returns_Impulse":
                    init_returns_impulse_found = True
                elif param.pname == "GetWave_Exists":
                    getwave_exists_found = True
            reserved_params_dict = tmp_params
        elif label == "Model_Specific":
            model_spec_found = True
            model_specific_dict = tmp_params
        elif label == "description":
            description = str(tmp_params)
        else:
            _err_str += f"WARNING: Unrecognized group with label, '{label}', found in parameter definition string!\n"

    if not reserved_found:
        _err_str += "ERROR: Reserved parameters section not found! It is required."

    if not init_returns_impulse_found:
        _err_str += "ERROR: Reserved parameter, 'Init_Returns_Impulse', not found! It is required."

    if not getwave_exists_found:
        _err_str += "ERROR: Reserved parameter, 'GetWave_Exists', not found! It is required."

    if not model_spec_found:
        _err_str += "WARNING: Model specific parameters section not found!"

    return (ParseErrMsg(_err_str), root_name, description, reserved_params_dict, model_specific_dict)


# Legacy client code support:
def parse_ami_param_defs(file_contents: str) -> tuple[ParseErrMsg, dict[str, Any]]:
    "The legacy version of ``parse_ami_file_contents()``."
    err_msg, root_name, description, reserved_params_dict, model_specific_dict = parse_ami_file_contents(file_contents)
    return (err_msg, {root_name: {"description":         description,
                                  "Reserved_Parameters": reserved_params_dict,
                                  "Model_Specific":      model_specific_dict}})


def make_gui(params: ModelSpecificDict) -> tuple[Group, list[TraitType]]:  # pylint: disable=too-many-locals
    """
    Builds top-level ``Group`` and list of ``Trait`` s from AMI parameter dictionary.

    Args:
        params: Dictionary of AMI parameters to be configured.

    Returns:
        A pair consisting of:

            - the top-level ``Group`` for the ``View``, and
            - a list of new ``Trait`` s created.

    Notes:
        1. The dictionary passed through ``params`` may have sub-dictionaries.
        The window layout will reflect this nesting.
        2. Some AMI files are written to encode the AMI parameter hierarchy
        via some parameter naming scheme. This function makes a primitive attempt
        to decode one particular example of this: ``<grp>_<param>``.
    """

    gui_items: list[Item | Group] = []
    new_traits: list[tuple[str, TraitType]] = []

    # Place parameters with a common prefix in their names,
    # but no pre-existing enclosing group, in a ``VGroup``.
    # - Generate grouped parameter names and list of unique group names.
    pnames = list(params.keys())
    pnames_split = list(map(lambda n: n.split('_'), pnames))
    group_names = list(map(lambda xs: xs[0], pnames_split))
    grouped_pnames = list(zip(pnames, group_names))
    # The following doesn't preserve the parameter ordering of the AMI file!
    # for group_name in set(group_names):  # Iterate over unique group names.
    unique_group_names = []
    for group_name in group_names:
        if group_name not in unique_group_names:
            unique_group_names.append(group_name)
    # - Process all parameters, trapping those from a non-explicit group for special processing.
    for group_name in unique_group_names:
        pnames_in_group = [pname for pname, grp_name in grouped_pnames if grp_name == group_name]
        if len(pnames_in_group) > 1:
            _gui_items: list[Item | Group] = []
            for pname in pnames_in_group:
                gui_item, new_trait = make_gui_items(pname, params[pname])
                _gui_items.extend(gui_item)
                new_traits.extend(new_trait)
            gui_items.append(VGroup(*_gui_items, label=group_name, show_border=True))
        else:
            pname = pnames_in_group[0]
            gui_item, new_trait = make_gui_items(pname, params[pname])
            gui_items.extend(gui_item)
            new_traits.extend(new_trait)

    # Finally, split into two rows if necessary.
    MAX_ITEMS_PER_ROW = 8
    if len(gui_items) > MAX_ITEMS_PER_ROW:
        rslt_grp = VGroup(
            HGroup(*(gui_items[:MAX_ITEMS_PER_ROW])),
            HGroup(*(gui_items[MAX_ITEMS_PER_ROW:])))
    else:
        rslt_grp = HGroup(*gui_items)

    return (rslt_grp, new_traits)


def make_gui_items(  # pylint: disable=too-many-locals,too-many-branches
    pname: str,
    param: AMIParameter | Parameters
) -> tuple[list[Item | Group], list[tuple[str, TraitType]]]:
    """
    Builds list of GUI items and list of traits from AMI parameter or dictionary.

    Args:
        pname: Parameter or sub-group name.
        param: AMI parameter or dictionary of AMI parameters to be configured.

    Returns:
        A pair consisting of:

            - the list of GUI items for the ``View``, and
            - the list of new ``Trait`` s created.

    Notes:
        1. A dictionary passed through ``param`` may have sub-dictionaries.
        These will be converted into sub- ``Group`` s in the returned list of GUI items.
    """

    if isinstance(param, AMIParameter):  # pylint: disable=no-else-return
        pusage = param.pusage
        if pusage not in ("In", "InOut"):
            return ([], [])

        if param.ptype == "Boolean":
            return ([Item(pname, tooltip=param.pdescription)], [(pname, Bool(param.pvalue))])

        pformat = param.pformat
        match pformat:
            case "Value":  # Value
                the_trait = Trait(param.pvalue)
            case "Range":
                the_trait = Range(param.pmin, param.pmax, param.pvalue)
            case "List":
                list_tips = param.plist_tip
                default = param.pdefault
                if list_tips:
                    tmp_dict: dict[str, Any] = {}
                    tmp_dict.update(list(zip(list_tips, param.pvalue)))
                    val = list(tmp_dict.keys())[0]
                    if default:
                        for tip in tmp_dict.items():
                            if tip[1] == default:
                                val = tip[0]
                                break
                    the_trait = Trait(val, tmp_dict)
                else:
                    val = default if default else param.pvalue[0]
                    the_trait = Enum([val] + param.pvalue)
            case "Corner":
                the_trait = Trait(param.pvalue[0])  # Using `Typ` value, for now.
            case _:
                raise ValueError(f"Unrecognized AMI parameter format: {pformat}, for parameter `{pname}` of type `{param.ptype}` and usage `{param.pusage}`!")
        if the_trait.metadata:
            the_trait.metadata.update({"transient": False})  # Required to support modal dialogs.
        else:
            the_trait.metadata = {"transient": False}
        return ([Item(name=pname, label=pname.split("_")[-1], tooltip=param.pdescription)], [(pname, the_trait)])

    else:  # subparameter branch
        gui_items: list[Item | Group] = []
        new_traits: list[tuple[str, TraitType]] = []
        subparam_names = list(param.keys())
        subparam_names.sort()
        group_desc = None

        # Build GUI items for this branch.
        for subparam_name in subparam_names:
            if subparam_name == "description":
                group_desc = param[subparam_name]
            else:
                tmp_items, tmp_traits = make_gui_items(pname + "_" + subparam_name, param[subparam_name])
                gui_items.extend(tmp_items)
                new_traits.extend(tmp_traits)

        if group_desc:
            gui_items = [Item(label=group_desc)] + gui_items

        return ([VGroup(*gui_items, label=pname.split("_")[-1], show_border=True)], new_traits)



#! /usr/bin/env python

"""Python tool for running several EmPy encoded tests on a IBIS-AMI model.

Original Author: David Banas
Original Date:   July 20, 2012

Copyright (c) 2012 David Banas; All rights reserved World wide.
"""

import shutil
from os import chdir
from pathlib import Path

import click
import em
import numpy as np

from pyibisami.ami.model    import AMIModel
from pyibisami.common       import TestSweep


def plot_name(tst_name, n=0):
    """Plot name generator keeps multiple tests from overwriting each other's
    plots."""
    while True:
        n += 1
        yield f"{tst_name}_plot_{n}.png"


def hsv2rgb(hue=0, saturation=1.0, value=1.0):
    """Convert a HSV number to and RGB one."""
    if value < 0:
        value = 0.0
    elif value > 1.0:
        value = 1.0
    if saturation == 0:
        return (value, value, value)
    if saturation < 0:
        saturation = 0
    elif saturation > 1.0:
        saturation = 1.0
    hue = hue % 360
    H = float(hue)
    S = float(saturation)
    V = float(value)
    H_i = np.floor(H / 60.0)
    f = (H / 60.0) - H_i
    p = V * (1.0 - S)
    q = V * (1.0 - f * S)
    t = V * (1.0 - (1.0 - f) * S)
    if H_i == 0:
        R = V
        G = t
        B = p
    elif H_i == 1:
        R = q
        G = V
        B = p
    elif H_i == 2:
        R = p
        G = V
        B = t
    elif H_i == 3:
        R = p
        G = q
        B = V
    elif H_i == 4:
        R = t
        G = p
        B = V
    else:
        R = V
        G = p
        B = q
    return (R, G, B)


def color_picker(num_hues=3, first_hue=0):
    """Yields pairs of colors having the same hue, but different intensities.

    The first color is fully bright and saturated, and the second is
    half bright and half saturated. Originally, the intent was to have
    the second color used for the `reference` waveform in plots.
    """
    hue = first_hue
    while True:
        yield (hsv2rgb(hue, 1.0, 1.0), hsv2rgb(hue, 0.75, 0.75))
        hue += 360 // num_hues


def expand_params(input_parameters: str) -> list[TestSweep]:
    """
    Take the command line input and convert it into usable parameter sweeps.

    We can pass in a file, directory, or raw string here.
    Handle all three cases.
    """
    if Path(input_parameters).exists():
        if Path(input_parameters).is_file():
            cfg_files = [Path(input_parameters)]
        else:
            cfg_files = list(Path(input_parameters).glob("*.run"))
        param_sweeps = []
        for cfg_filename in cfg_files:
            cfg_name = cfg_filename.stem
            param_list = []
            with open(cfg_filename, "rt", encoding="utf-8") as cfg_file:
                description = cfg_file.readline()
                expr = ""
                for line in cfg_file:
                    toks = line.split()
                    if not toks or toks[0].startswith("#"):
                        continue
                    expr += line
                    if toks[-1] == "\\":  # Test for line continuation.
                        expr = expr.rstrip("\\\n")
                    else:
                        param_list.append(eval(compile(expr, cfg_filename, "eval")))  # pylint: disable=eval-used
                        expr = ""
            param_sweeps.append((cfg_name, description, param_list))
    else:
        try:
            param_sweeps = eval(input_parameters)  # pylint: disable=eval-used
        except Exception as err:
            raise ValueError(f"input_parameters: {input_parameters}") from err
    return param_sweeps


def run_tests(**kwargs):  # pylint: disable=too-many-locals
    """Provide a thin wrapper around the click interface so that we can test
    the operation."""

    # Fetch options and cast into local independent variables.
    test_dir = Path(kwargs["test_dir"]).resolve()
    ref_dir = Path(kwargs["ref_dir"]).resolve()
    if not ref_dir.exists():
        ref_dir = None
    model = Path(kwargs["model"]).resolve()
    out_dir = Path(kwargs["out_dir"]).resolve()
    out_dir.mkdir(exist_ok=True)
    xml_filename = out_dir.joinpath(kwargs["xml_file"])

    # Some browsers demand that the stylesheet be located in the same
    # folder as the *.XML file. Besides, this allows the model tester
    # to zip up her 'test_results' directory and send it off to
    # someone, whom may not have the PyIBIS-AMI package installed.
    #
    # Note: To avoid this issue entirely, incorporate `xsltproc` into your build flow.
    shutil.copy(str(Path(__file__).parent.joinpath("test_results.xsl")), str(out_dir))

    print(f"Testing model: {model}")
    print(f"Using tests in: {test_dir}")
    params = expand_params(kwargs["params"])

    # Run the tests.
    print(f"Sending XHTML output to: {xml_filename}")
    with open(xml_filename, "w", encoding="utf-8") as xml_file:
        xml_file.write('<?xml version="1.0" encoding="ISO-8859-1"?>\n')
        xml_file.write('<?xml-stylesheet type="text/xsl" href="test_results.xsl"?>\n')
        xml_file.write("<tests>\n")
    if kwargs["tests"]:
        tests = kwargs["tests"]
    else:
        tests = list(test_dir.glob("*.em"))
    for test in tests:
        test_ = test.stem
        print(f"Running test: {test_} ...")
        theModel = AMIModel(str(model))
        plot_names = plot_name(test_)
        for cfg_item in params:
            cfg_name = cfg_item[0]
            print(f"\tRunning test configuration: {cfg_name} ...")
            description = cfg_item[1]
            param_list = cfg_item[2]
            colors = color_picker(num_hues=len(param_list))
            with open(xml_filename, "a", encoding="utf-8") as xml_file:
                interpreter = em.Interpreter(
                    output=xml_file,
                    globals={
                        "name": f"{test_} ({cfg_name})",
                        "model": theModel,
                        "data": param_list,
                        "plot_names": plot_names,
                        "description": description,
                        "plot_colors": colors,
                        "ref_dir": ref_dir,
                    },
                )
                try:
                    cwd = Path().cwd()
                    chdir(out_dir)  # So that the images are saved in the output directory.
                    with open(Path(test_dir, test), encoding="utf-8") as test_file:
                        interpreter.file(test_file)
                    chdir(cwd)
                except Exception as err:  # pylint: disable=broad-exception-caught
                    print("\t\t", err)
                finally:
                    interpreter.shutdown()
        print("Test:", test_, "complete.")
    with open(xml_filename, "a", encoding="utf-8") as xml_file:
        xml_file.write("</tests>\n")

    print(f"Please, open file, `{xml_filename}` in a Web browser, in order to view the test results.")


@click.command(context_settings={"ignore_unknown_options": True, "help_option_names": ["-h", "--help"]})
@click.option(
    "--model", "-m", default="libami.so", type=click.Path(exists=True), help="Sets the AMI model DLL file name."
)
@click.option(
    "--test_dir",
    "-t",
    default="tests",
    type=click.Path(),
    help="Sets the name of the directory from which tests are taken.",
)
@click.option(
    "--params",
    "-p",
    default='[("cfg_dflt", "default", [("default", ({"root_name":"testAMI"},{})),]),]',
    help='List of lists of model configurations. Format: <filename> or [(name, [(label, ({AMI params., in "key:val" format},{Model params., in "key:val" format})), ...]), ...]',
)
@click.option(
    "--xml_file",
    "-x",
    default="test_results.xml",
    help="Sets the name of the XML output file. You should load this file into your Web browser after the program completion.",
)
@click.option(
    "--ref_dir",
    "-r",
    default="refs",
    type=click.Path(),
    help="Sets the name of the directory from which reference waveforms are taken.",
)
@click.option(
    "--out_dir",
    "-o",
    default="test_results",
    type=click.Path(),
    help="Sets the name of the directory in which to place the results.",
)
@click.argument("tests", nargs=-1, type=click.UNPROCESSED)
@click.version_option()
def main(**kwargs):
    """Run a series of tests on a AMI model DLL file.

    If no tests are specified on the command line, run all tests found
    in `test_dir'. (See `-t' option.)

    usage: %prog [options] [test1 [test2 ...]]

    Tests are written in the EmPy templating language, and produce XML
    output. (See the examples provided in the `examples' directory of the
    `pyibisami' Python package.)

    Test results should be viewed by loading the XML output file into
    a Web browser. By default, the XML output file refers to the supplied
    XSLT file, `test_results.xsl'. It is possible that you may need to
    copy this file from the pyibisami package directory to your local
    working directory, in order to avoid file loading errors in your
    Web browser.

    If your browser's security settings disallow the use of an XSLT style
    sheet when processing XML then use the `xsltproc` utility to convert
    the XML/XSLT to HTML before viewing in your browser.
    """
    run_tests(**kwargs)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



from pathlib import Path

import pytest

from pyibisami.tools.run_tests import (
    color_picker,
    expand_params,
    hsv2rgb,
    plot_name,
    run_tests,
)


class TestRunTests(object):
    def test_hsv2rgb(self):
        """Convert an HSV number to a RGB one. Everything is normalized to 1."""
        # Black
        assert hsv2rgb(0, 0, 0) == (0.0, 0.0, 0.0)
        # Cyan
        assert hsv2rgb(180, 1, 1) == (0.0, 1.0, 1.0)
        # Red
        assert hsv2rgb(0, 1, 1) == (1.0, 0.0, 0.0)
        # Lime
        assert hsv2rgb(120, 1, 1) == (0.0, 1.0, 0.0)
        # Blue
        assert hsv2rgb(240, 1, 1) == (0.0, 0.0, 1.0)

    def test_plot_name(self):
        image_filename = plot_name("pytest")
        assert next(image_filename) == "pytest_plot_1.png"

    def test_color_picker(self):
        color = color_picker()
        assert next(color) == (hsv2rgb(0, 1.0, 1.0), hsv2rgb(0, 0.75, 0.75))

    def test_expand_params(self):
        input_params = '[("cfg_dflt", "default", [("default", ({"root_name":"testAMI"},{})),]),]'
        params = expand_params(input_params)
        assert params == [("cfg_dflt", "default", [("default", ({"root_name": "testAMI"}, {}))])]

    def test_expand_params_file(self, tmp_path):
        run_file = tmp_path.joinpath("test.run")
        with open(run_file, "w") as test_file:
            test_file.write(
                """Sweep of Tx pre-emphasis filter posttap1 values.

('Posttap1 = 0', \
  ({'root_name'        : 'example_tx', \
    'tx_tap_np1'       : 0, \
    'tx_tap_nm1'       : 0, \
    'tx_tap_nm2'       : 0, \
   }, \
   {'channel_response' : [2.0e11] + [0.], \
    'sample_interval'  : 5.0e-12, \
   }, \
  ), \
)"""
            )
        params = expand_params(run_file)
        assert params == [
            (
                "test",
                "Sweep of Tx pre-emphasis filter posttap1 values.\n",
                [
                    (
                        "Posttap1 = 0",
                        (
                            {
                                "root_name": "example_tx",
                                "tx_tap_nm1": 0,
                                "tx_tap_nm2": 0,
                                "tx_tap_np1": 0,
                            },
                            {"channel_response": [200000000000.0, 0.0], "sample_interval": 5e-12},
                        ),
                    )
                ],
            )
        ]

    def test_expand_params_directory(self, tmp_path):
        runs_dir = tmp_path.joinpath("runs")
        runs_dir.mkdir()
        run_file = runs_dir.joinpath("test.run")
        with open(run_file, "w") as test_file:
            test_file.write(
                """Sweep of Tx pre-emphasis filter posttap1 values.

('Posttap1 = 0', \
  ({'root_name'        : 'example_tx', \
    'tx_tap_np1'       : 0, \
    'tx_tap_nm1'       : 0, \
    'tx_tap_nm2'       : 0, \
   }, \
   {'channel_response' : [2.0e11] + [0.], \
    'sample_interval'  : 5.0e-12, \
   }, \
  ), \
)"""
            )
        params = expand_params(runs_dir)
        assert params == [
            (
                "test",
                "Sweep of Tx pre-emphasis filter posttap1 values.\n",
                [
                    (
                        "Posttap1 = 0",
                        (
                            {
                                "root_name": "example_tx",
                                "tx_tap_nm1": 0,
                                "tx_tap_nm2": 0,
                                "tx_tap_np1": 0,
                            },
                            {"channel_response": [200000000000.0, 0.0], "sample_interval": 5e-12},
                        ),
                    )
                ],
            )
        ]

    @pytest.mark.xfail(reason="EMPY looses its stdout proxy.")
    def test_run_tests(self):
        model = Path(__file__).parents[1].joinpath("examples", "example_tx_x86_amd64.so")
        test_dir = Path(__file__).parents[1].joinpath("examples", "tests")
        params = Path(__file__).parents[1].joinpath("examples", "runs")
        xml_file = "test_results.xml"
        ref_dir = Path().cwd()
        out_dir = Path(__file__).parents[1].joinpath("examples", "test_results")
        run_tests(
            model=model,
            test_dir=test_dir,
            params=params,
            xml_file=xml_file,
            ref_dir=ref_dir,
            out_dir=out_dir,
            tests=(),
        )



"""
Class definitions for working with IBIS-AMI models.

Original Author: David Banas

Original Date:   July 3, 2012

Copyright (c) 2019 David Banas; All rights reserved World wide.
"""

import copy as cp
from ctypes import CDLL, byref, c_char_p, c_double  # pylint: disable=no-name-in-module
from pathlib import Path
from typing import Any, Optional

import numpy as np
from numpy.random     import default_rng

from pyibisami.common import Rvec, deconv_same


def loadWave(filename: str) -> tuple[Rvec, Rvec]:
    """
    Load a waveform file.

    The file should consist of any number of lines, where each line
    contains, first, a time value and, second, a voltage value.
    Assume the first line is a header, and discard it.

    Specifically, this function may be used to load in waveform files
    saved from *CosmosScope*.

    Args:
        filename: Name of waveform file to read in.

    Returns:
        A pair of *NumPy* arrays containing the time and voltage values, respectively.
    """

    with open(filename, "r", encoding="utf-8") as theFile:
        theFile.readline()  # Consume the header line.
        time = []
        voltage = []
        for line in theFile:
            tmp = list(map(float, line.split()))
            time.append(tmp[0])
            voltage.append(tmp[1])
        return (np.array(time), np.array(voltage))


def interpFile(filename: str, sample_per: float) -> Rvec:
    """
    Read in a waveform from a file, and convert it to the given sample rate,
    using linear interpolation.

    Args:
        filename: Name of waveform file to read in.
        sample_per: New sample interval, in seconds.

    Returns:
        A *NumPy* array containing the resampled waveform.
    """

    impulse = loadWave(filename)
    ts = impulse[0]
    ts = ts - ts[0]
    vs = impulse[1]
    tmax = ts[-1]
    # Build new impulse response, at new sampling period, using linear interpolation.
    res = []
    t = 0.0
    i = 0
    while t < tmax:
        while ts[i] <= t:
            i = i + 1
        res.append(vs[i - 1] + (vs[i] - vs[i - 1]) * (t - ts[i - 1]) / (ts[i] - ts[i - 1]))
        t = t + sample_per
    return np.array(res)


class AMIModelInitializer:
    """
    Class containing the initialization data for an instance of ``AMIModel``.

    Created primarily to facilitate use of the PyAMI package at the
    pylab command prompt, this class can be used by the pylab user, in
    order to store all the data required to initialize an instance of
    class ``AMIModel``. In this way, the pylab user may assemble the
    AMIModel initialization data just once, and modify it incrementally,
    as she experiments with different initialization settings. In this
    way, she can avoid having to type a lot of redundant constants every
    time she invokes the AMIModel constructor.
    """

    # pylint: disable=too-few-public-methods,too-many-instance-attributes

    ami_params = {"root_name": ""}

    _init_data = {
        "channel_response": (c_double * 128)(0.0, 1.0, 0.0),
        "row_size": 128,
        "num_aggressors": 0,
        "sample_interval": c_double(25.0e-12),
        "bit_time": c_double(0.1e-9),
    }

    def __init__(self, ami_params: dict, info_params: Optional[dict] = None, **optional_args):
        """
        Constructor accepts a mandatory dictionary containing the AMI
        parameters, as well as optional information parameter dictionary
        and initialization data overrides, and validates them, before
        using them to update the local initialization data structures.

        Valid names of optional initialization data overrides:

        - channel_response
            a matrix of ``c_double's`` where the first row represents the
            impulse response of the analog channel, and the rest represent
            the impulse responses of several aggressor-to-victim far end
            crosstalk (FEXT) channels.

            Default) a single 128 element vector containing an ideal impulse

        - row_size
            integer giving the size of the rows in ``channel_response``.

            Default) 128

        - num_aggressors
            integer giving the number or rows in ``channel_response``, minus
            one.

            Default) 0

        - sample_interval
            c_double giving the time interval, in seconds, between
            successive elements in any row of ``channel_response``.

            Default) 25e-12 (40 GHz sampling rate)

        - bit_time
            c_double giving the bit period (i.e. - unit interval) of the
            link, in seconds.

            Default) 100e-12 (10 Gbits/s)
        """

        self.ami_params = {"root_name": ""}
        self.ami_params.update(ami_params)
        self.info_params = info_params

        # Need to reverse sort, in order to catch ``sample_interval`` and ``row_size``,
        # before ``channel_response``, since ``channel_response`` depends upon ``sample_interval``,
        # when ``h`` is a file name, and overwrites ``row_size``, in any case.
        keys = list(optional_args.keys())
        keys.sort(reverse=True)
        if keys:
            for key in keys:
                if key in self._init_data:
                    self._init_data[key] = optional_args[key]

    def __str__(self):
        return "\n\t".join([
            "AMIModelInitializer instance:",
            f"`ami_params`: {self.ami_params}",
            f"`info_params`: {self.ami_params}"])

    def _getChannelResponse(self):
        return list(map(float, self._init_data["channel_response"]))

    def _setChannelResponse(self, h):
        if isinstance(h, str) and Path(h).is_file():
            h = interpFile(h, self.sample_interval)
        Vector = c_double * len(h)
        self._init_data["channel_response"] = Vector(*h)
        self.row_size = len(h)

    channel_response = property(
        _getChannelResponse,
        _setChannelResponse,
        doc="Channel impulse response to be passed to AMI_Init(). May be a file name.",
    )

    def _getRowSize(self):
        return self._init_data["row_size"]

    def _setRowSize(self, n):
        self._init_data["row_size"] = n

    row_size = property(_getRowSize, _setRowSize, doc="Number of elements in channel response vector(s).")

    def _getNumAggressors(self):
        return self._init_data["num_aggressors"]

    def _setNumAggressors(self, n):
        self._init_data["num_aggressors"] = n

    num_aggressors = property(
        _getNumAggressors, _setNumAggressors, doc="Number of vectors in ``channel_response``, minus one."
    )

    def _getSampleInterval(self):
        return float(self._init_data["sample_interval"].value)

    def _setSampleInterval(self, T):
        self._init_data["sample_interval"] = c_double(T)

    sample_interval = property(
        _getSampleInterval,
        _setSampleInterval,
        doc="Time interval between adjacent elements in channel response vector(s).",
    )

    def _getBitTime(self):
        return float(self._init_data["bit_time"].value)

    def _setBitTime(self, T):
        self._init_data["bit_time"] = c_double(T)

    bit_time = property(_getBitTime, _setBitTime, doc="Link unit interval.")


class AMIModel:  # pylint: disable=too-many-instance-attributes
    """
    Class defining the structure and behavior of an IBIS-AMI Model.

    Notes:
        1. Makes the calling of ``AMI_Close()`` automagic,
        by calling it from the destructor.
    """

    def __init__(self, filename: str):
        """
        Load the dll and bind the 3 AMI functions.

        Args:
            filename: The DLL/SO file name.

        Raises:
            OSError: If given file cannot be opened.
        """

        self._filename = filename
        self._ami_mem_handle = None
        my_dll = CDLL(filename)
        self._amiInit = my_dll.AMI_Init
        self._amiClose = my_dll.AMI_Close
        try:
            self._amiGetWave = my_dll.AMI_GetWave
        except Exception:  # pylint: disable=broad-exception-caught
            self._amiGetWave = None  # type: ignore

    def __del__(self):
        """
        Destructor - Calls ``AMI_Close()`` with handle to AMI model memory.

        This obviates the need for the user to call the ``AMI_Close()``
        function explicitly, and guards against memory leaks, during
        PyLab command prompt operation, by ensuring that ``AMI_Close()``
        gets called automagically when the model goes out of scope.
        """
        if self._ami_mem_handle:
            self._amiClose(self._ami_mem_handle)

    def __str__(self):
        return "\n\t".join([
            f"AMIModel instance: `{self._filename}`",
            f"Length of initOut = {len(self._initOut)}",
            f"row_size = {self._row_size}",
            f"num_aggressors = {self._num_aggressors}",
            f"sample_interval = {self._sample_interval}",
            f"bit_time = {self._bit_time}",
            f"samps_per_bit = {self._samps_per_bit}",
            f"bits_per_call = {self._bits_per_call}",
            f"ami_params_in = {self._ami_params_in}",
            f"ami_params_out = {self._ami_params_out}",
            f"&ami_mem_handle = {byref(self._ami_mem_handle)}",
            f"Message = {self._msg}",
            f"AMI_Init(): {self._amiInit}",
            f"AMI_GetWave(): {self._amiGetWave}",
            f"AMI_Close(): {self._amiClose}",])

    def initialize(self, init_object: AMIModelInitializer):
        """
        Wraps the ``AMI_Init()`` function.

        Args:
            init_object: The model initialization data.

        Notes:
            1. Takes an instance of ``AMIModelInitializer`` as its only argument.
            This allows model initialization data to be constructed once,
            and modified incrementally in between multiple calls of
            ``initialize``. This is useful for *PyLab* command prompt testing.

        ToDo:
            1. Allow for non-integral number of samples per unit interval.
        """

        # Free any memory allocated by the previous initialization.
        if self._ami_mem_handle:
            self._amiClose(self._ami_mem_handle)

        # Set up the AMI_Init() arguments.
        self._channel_response = (   # pylint: disable=attribute-defined-outside-init
            init_object._init_data[  # pylint: disable=protected-access
                "channel_response"
            ]
        )
        self._initOut = cp.copy(self._channel_response)  # type: ignore  # pylint: disable=attribute-defined-outside-init
        self._row_size = init_object._init_data[  # pylint: disable=protected-access,attribute-defined-outside-init
            "row_size"
        ]
        self._num_aggressors = init_object._init_data[  # pylint: disable=protected-access,attribute-defined-outside-init
            "num_aggressors"
        ]
        self._sample_interval = (    # pylint: disable=attribute-defined-outside-init
            init_object._init_data[  # pylint: disable=protected-access
                "sample_interval"
            ]
        )
        self._bit_time = init_object._init_data[  # pylint: disable=protected-access,attribute-defined-outside-init
            "bit_time"
        ]
        self._info_params = init_object.info_params  # pylint: disable=attribute-defined-outside-init

        # Check GetWave() consistency if possible.
        if init_object.info_params and init_object.info_params["GetWave_Exists"] and init_object.info_params["GetWave_Exists"].pvalue:
            if not self._amiGetWave:
                raise RuntimeError(
                    "Reserved parameter `GetWave_Exists` is True, but I can't bind to `AMI_GetWave()`!"
                )

        # Construct the AMI parameters string.
        def sexpr(pname: str, pval: Any) -> str:
            """
            Create an S-expression from a parameter name/value pair, calling
            recursively as needed to elaborate sub-parameter dictionaries.

            Args:
                pname: Parameter name.
                pval: Parameter value.

            Returns:
                A string containing the S-expression constructed.
            """
            if isinstance(pval, str):
                return f'({pname} "{pval}")'
            if isinstance(pval, dict):
                subs = []
                for sname in pval:
                    subs.append(sexpr(sname, pval[sname]))
                # return sexpr(pname, " ".join(subs))
                return f"({pname} {' '.join(subs)})"
            return f"({pname} {pval})"

        root_name = init_object.ami_params['root_name']
        self._root_name = root_name  # pylint: disable=attribute-defined-outside-init
        ami_params_in = f"({root_name} "
        for item in list(init_object.ami_params.items()):
            if not item[0] == "root_name":
                ami_params_in += sexpr(item[0], item[1])
        ami_params_in += ")"
        self._ami_params_in = ami_params_in.encode("utf-8")  # pylint: disable=attribute-defined-outside-init

        # Set handle types.
        self._ami_params_out = c_char_p(b"")  # pylint: disable=attribute-defined-outside-init
        self._ami_mem_handle = c_char_p(None)  # type: ignore  # pylint: disable=attribute-defined-outside-init
        self._msg = c_char_p(b"")  # pylint: disable=attribute-defined-outside-init

        # Call AMI_Init(), via our Python wrapper.
        try:
            self._amiInit(
                byref(self._initOut),  # type: ignore
                self._row_size,
                self._num_aggressors,
                self._sample_interval,
                self._bit_time,
                self._ami_params_in,  # Prevents model from mucking up our input parameter string.
                byref(self._ami_params_out),
                byref(self._ami_mem_handle),  # type: ignore
                byref(self._msg),
            )
        except OSError as err:
            print("pyibisami.ami_model.AMIModel.initialize(): Call to AMI_Init() bombed:")
            print(err)
            print(f"AMI_Init() address = {self._amiInit}")
            print("Values sent into AMI_Init():")
            print(f"&initOut = {byref(self._initOut)}")  # type: ignore
            print(f"row_size = {self._row_size}")
            print(f"num_aggressors = {self._num_aggressors}")
            print(f"sample_interval = {self._sample_interval}")
            print(f"bit_time = {self._bit_time}")
            print(f"ami_params_in = {ami_params_in}")
            print(f"&ami_params_out = {byref(self._ami_params_out)}")
            print(f"&ami_mem_handle = {byref(self._ami_mem_handle)}")  # type: ignore
            print(f"&msg = {byref(self._msg)}")
            raise err

        # Initialize attributes used by getWave().
        bit_time = init_object.bit_time
        sample_interval = init_object.sample_interval
        # ToDo: Fix this. There isn't actually a requirement that `bit_time` be an integral multiple of `sample_interval`.
        # And there may be an advantage to having it not be!
        # if (bit_time % sample_interval) > (sample_interval / 100):
        #     raise ValueError(
        #         f"Bit time ({bit_time * 1e9: 6.3G} ns) must be an integral multiple of sample interval ({sample_interval * 1e9: 6.3G} ns)."
        #     )
        self._samps_per_bit = int(bit_time / sample_interval)  # pylint: disable=attribute-defined-outside-init
        self._bits_per_call = (  # pylint: disable=attribute-defined-outside-init
            init_object.row_size / self._samps_per_bit
        )

    def getWave(self, wave: Rvec, bits_per_call: int = 0) -> tuple[Rvec, Rvec, list[str]]:  # noqa: F405
        """
        Performs time domain processing of input waveform, using the ``AMI_GetWave()`` function.

        Args:
            wave: Waveform to be processed.

        Keyword Args:
            bits_per_call: Number of bits to use, per call to ``AMI_GetWave()``.
                Default: 0 (Means "Use existing value.")

        Returns:
            A tuple containing

                - the processed waveform,
                - the recovered slicer sampling instants, and
                - the list of output parameter strings received from each call to ``AMI_GetWave()``.

        Notes:
            1. The returned clock times are given in "pre-edge-aligned" fashion,
            which means their values are: sampling instant - ui/2.
        """

        if bits_per_call:
            self._bits_per_call = int(bits_per_call)  # pylint: disable=attribute-defined-outside-init
        bits_per_call = int(self._bits_per_call)
        samps_per_call = int(self._samps_per_bit * bits_per_call)

        # Create the required C types.
        Signal = c_double * samps_per_call
        Clocks = c_double * (bits_per_call + 1)  # The "+1" is critical, to prevent access violations by the model.

        idx = 0  # Holds the starting index of the next processing chunk.
        _clock_times = Clocks(0.0)
        wave_out: list[float] = []
        clock_times: list[float] = []
        params_out: list[str] = []
        input_len = len(wave)
        while idx < input_len:
            remaining_samps = input_len - idx
            if remaining_samps < samps_per_call:
                Signal = c_double * remaining_samps
                tmp_wave = wave[idx:]
            else:
                tmp_wave = wave[idx: idx + samps_per_call]
            _wave = Signal(*tmp_wave)
            try:
                self._amiGetWave(
                    byref(_wave), len(_wave), byref(_clock_times),
                    byref(self._ami_params_out), self._ami_mem_handle
                )  # type: ignore
            except OSError:
                print(self)
                print(f"byref(_wave): {byref(_wave)}")
                print(f"len(_wave): {len(_wave)}")
                print(f"byref(_clock_times): {byref(_clock_times)}")
                print(f"byref(self._ami_params_out): {byref(self._ami_params_out)}")
                print(f"self._ami_mem_handle: {self._ami_mem_handle}")
                raise
            wave_out.extend(_wave)
            clock_times.extend(_clock_times)
            params_out.append(self.ami_params_out)
            idx += len(_wave)

        return np.array(wave_out), np.array(clock_times[: len(wave_out) // self._samps_per_bit]), params_out

    def get_responses(  # pylint: disable=too-many-locals
        self,
        bits_per_call: int = 0,
        pad_bits: int = 10,
        nbits: int = 200,
        calc_getw: bool = True
    ) -> dict[str, Any]:
        """
        Get the impulse response of an initialized IBIS-AMI model, alone and convolved with the channel.

        Keyword Args:
            bits_per_call: Number of bits to include in the input to `GetWave()`.
                Default: 0 (Means "use model's existing value".)
            pad_bits: Number of bits to pad leading edge with when calling `GetWave()`,
                to protect from initial garbage in `GetWave()` output.
                Default: 10
            nbits: Number of "real" bits to use for `GetWave()` testing.
                Default: 200
            calc_getw: Calculate ``GetWave()`` responses, also, when True.
                Default: True

        Returns:
            Dictionary containing the responses under the following keys

                - "imp_resp_init": The model's impulse response, from its `AMI_Init()` function (V/sample).
                - "out_resp_init": `imp_resp_init` convolved with the channel.
                - "imp_resp_getw": The model's impulse response, from its `AMI_GetWave()` function (V/sample).
                - "out_resp_getw": `imp_resp_getw` convolved with the channel.

        Notes:
            1. If either set of keys (i.e. - "..._init" or "..._getw")
            is missing from the returned dictionary, it means that
            that mode of operation (`AMI_Init()` or `AMI_GetWave()`)
            was not available in the given model.

            2. An empty dictionary implies that neither the `Init_Returns_Impulse`
            nor the `GetWave_Exists` AMI reserved parameter was True.

            3. Note that impulse responses are returned with units: (V/sample), not (V/s).

        ToDo:
            1. Implement `bit_gen`.
            2. Implement `ignore_bits`.
        """

        rslt = {}

        # Capture needed parameter definitions.
        ui = self.bit_time
        ts = self.sample_interval
        info_params = self.info_params
        ignore_bits = info_params["Ignore_Bits"].pvalue if "Ignore_Bits" in info_params else 0

        # Capture/convert instance variables.
        chnl_imp = np.array(self.channel_response) * ts     # input (a.k.a. - "channel") impulse response (V/sample)
        out_imp = np.array(self.initOut) * ts               # output impulse response (V/sample)

        # Calculate some needed intermediate values.
        nspui = int(ui / ts)            # samps per UI
        pad_samps = pad_bits * nspui    # leading edge padding samples for GetWave() calls
        len_h = len(out_imp)
        t = np.array([i * ts for i in range(-pad_samps, len_h - pad_samps)])
        f = np.array([i * 1.0 / (ts * len_h) for i in range(len_h // 2 + 1)])  # Assumes `rfft()` is used.

        # Extract and return the model responses.
        if self.info_params["Init_Returns_Impulse"]:
            h_model = deconv_same(out_imp, chnl_imp)  # noqa: F405
            rslt["imp_resp_init"] = np.roll(h_model, -len(h_model) // 2 + 3 * nspui)

            h_init = np.roll(out_imp, pad_samps)
            s_init = np.cumsum(h_init)                 # Step response.
            p_init = s_init - np.pad(s_init[:-nspui], (nspui, 0), mode='constant', constant_values=0)
            H_init = np.fft.rfft(self.initOut)
            H_init *= s_init[-1] / np.abs(H_init[0])   # Normalize for proper d.c.
            rslt["out_resp_init"] = (t, h_init, s_init, p_init, f, H_init)

        if calc_getw and self.info_params["GetWave_Exists"].pvalue:
            # Get model's step response.
            rng = default_rng()
            u = np.concatenate(
                (rng.integers(low=0, high=2, size=ignore_bits),
                 np.array([0] * pad_bits + [1] * nbits))).repeat(nspui) - 0.5
            wave_out, _, _ = self.getWave(u, bits_per_call=bits_per_call)

            # Calculate impulse response from step response.
            rslt["imp_resp_getw"] = np.diff(wave_out[(ignore_bits + pad_bits) * nspui:])

            # Get step response of channel + model.
            wave_in = np.convolve(u, chnl_imp)[:len(u)]
            wave_out, _, _ = self.getWave(wave_in, bits_per_call=bits_per_call)
            s_getw = wave_out[ignore_bits * nspui:][:len(t)] + 0.5
            # Match the d.c. offset of Init() output, for easier comparison of Init() & GetWave() outputs.
            s_getw -= s_getw[pad_samps - 1]
            p_getw = s_getw - np.pad(s_getw[:-nspui], (nspui, 0), mode='constant', constant_values=0)
            _s = s_getw[pad_samps:]
            h_getw = np.insert(np.diff(_s), 0, _s[0])
            len_hgw = len(h_getw)
            if len_hgw > len_h:
                h_getw = h_getw[:len_h]
            else:
                h_getw = np.pad(h_getw, (0, len_h - len_hgw))
            H_getw = np.fft.rfft(h_getw)
            rslt["out_resp_getw"] = (t, h_getw, s_getw, p_getw, f, H_getw)

        return rslt

    @property
    def root_name(self):
        """AMI parameter tree root name."""
        return self._root_name

    def _getInitOut(self):
        return list(map(float, self._initOut))

    initOut = property(_getInitOut, doc="Channel response convolved with model impulse response.")

    def _getChannelResponse(self):
        return list(map(float, self._channel_response))

    channel_response = property(_getChannelResponse, doc="Channel response passed to initialize().")

    def _getRowSize(self):
        return self._row_size

    row_size = property(_getRowSize, doc="Length of vector(s) passed to initialize().")

    def _getNumAggressors(self):
        return self._num_aggressors

    num_aggressors = property(_getNumAggressors, doc="Number of rows in matrix passed to initialize(), minus one.")

    def _getSampleInterval(self):
        return float(self._sample_interval.value)

    sample_interval = property(
        _getSampleInterval, doc="Time interval between adjacent elements of the vector(s) passed to initialize()."
    )

    def _getBitTime(self):
        return float(self._bit_time.value)

    bit_time = property(_getBitTime, doc="Link unit interval, as passed to initialize().")

    def _getAmiParamsIn(self):
        return self._ami_params_in

    ami_params_in = property(_getAmiParamsIn, doc="The AMI parameter string passed to AMI_Init() by initialize().")

    def _getAmiParamsOut(self):
        pout: str = self._ami_params_out.value
        if pout:  # pylint: disable=using-constant-test
            return pout.decode()  # pylint: disable=no-member
        return f"({self.root_name} )"

    ami_params_out = property(
        _getAmiParamsOut, doc="The AMI parameter string returned by either `AMI_Init()` or `AMI_GetWave()`."
    )

    def _getMsg(self):
        return self._msg.value

    msg = property(_getMsg, doc="Message returned by most recent call to AMI_Init() or AMI_GetWave().")

    def _getClockTimes(self):
        return self.clock_times

    clock_times = property(_getClockTimes, doc="Clock times returned by most recent call to getWave().")

    def _getInfoParams(self):
        return self._info_params

    info_params = property(_getInfoParams, doc="Reserved AMI parameter values for this model.")



#! /usr/bin/env python

"""
Run an IBIS-AMI model through a Jupyter notebook, generating an HTML report.

Original Author: David Banas

Original Date:   January 31, 2025

Copyright (c) 2025 David Banas; all rights reserved World wide.
"""

import os
from pathlib    import Path
import subprocess
import sys
import shlex
from time       import time
from typing     import Any, Optional

import click

NOTEBOOK = Path(__file__).parent.parent.joinpath("IBIS_AMI_Tester.ipynb")


def run_notebook(
    ibis_file: Path,
    notebook: Path,
    out_dir: Optional[Path] = None,
    notebook_params: Optional[dict[str, Any]] = None,
) -> None:
    """
    Run a Jupyter notebook on the target IBIS-AMI model.

    Args:
        ibis_file: The ``*.ibs`` file to test.
            (Presumably, an IBIS-AMI model.)
        notebook: The *Jupyter* notebook to use for testing the model.

    Keyword Args:
        out_dir: The directory into which to place the resultant HTML file.
            Default: None (Use the directory containing the ``*.ibs`` file.)
        notebook_params: An optional dictionary of parameter overrides for the notebook.
            Default: None
    """

    start_time = time()

    # Validate input.
    if not ibis_file.exists():
        raise RuntimeError(f"Can't find IBIS-AMI model file, {ibis_file}!")
    if not notebook.exists():
        raise RuntimeError(f"Can't find notebook file, {notebook}!")

    # Define temp. (i.e. - parameterized) notebook and output file locations.
    tmp_dir = (
        os.environ.get("TMP") or os.environ.get("TEMP") or  # noqa: W504
        (Path(os.environ.get("HOME")).joinpath("tmp")  # type: ignore
            if os.environ.get("HOME")
            else "/tmp")
    )
    tmp_dir = Path(tmp_dir)
    tmp_dir.mkdir(exist_ok=True)
    tmp_notebook = tmp_dir.joinpath(notebook.stem + "_papermill").with_suffix(".ipynb")
    out_dir = out_dir or ibis_file.resolve().parent
    out_dir.mkdir(exist_ok=True)
    html_file = Path(out_dir.joinpath(ibis_file.name)).with_suffix(".html")

    # Run the notebook.
    print(f"Testing IBIS-AMI model: {ibis_file},")
    print(f"using notebook: {notebook},")
    print(f"sending HTML output to: {html_file}...")

    try:
        extra_args = []
        if notebook_params:
            extra_args = [tok for item in notebook_params.items()
                              for tok  in ['-p', f'{item[0]}', f'{item[1]}']]  # noqa: E127
        subprocess.run(['papermill', str(notebook), str(tmp_notebook)] + extra_args, check=True)
    except Exception:
        print(f"notebook: {notebook}")
        print(f"tmp_notebook: {tmp_notebook}")
        raise
    subprocess.run(
        ['jupyter', 'nbconvert', '--to', 'html', '--no-input', '--output', html_file, tmp_notebook],
        check=True)

    run_time = int(time() - start_time)  # integer seconds
    hours, rem_secs  = divmod(run_time, 3600)
    minutes, seconds = divmod(rem_secs, 60)
    print(f"Done after {hours} hrs., {minutes} min., {seconds} sec.")


@click.command(context_settings={"ignore_unknown_options": False,
                                 "help_option_names": ["-h", "--help"]})
@click.option(
    "--notebook", "-n", default=NOTEBOOK, type=click.Path(exists=True),
    help="Override the default notebook file name."
)
@click.option(
    "--out-dir", "-o", default=None, type=click.Path(),
    help="Override the name of the directory in which to place the results."
)
@click.option(
    "--params", "-p",
    default='',
    help='Directory (or, file) containing configuration sweeps.',
)
@click.option("--debug", is_flag=True, help="Provide extra debugging information.")
@click.option("--is_tx", is_flag=True, help="Flags a Tx model.")
@click.option("--nspui", default=32, show_default=True, help="Number of samples per unit interval.")
@click.option("--no-nspui-swp", is_flag=True, help="Skip samples per UI sweep.")
@click.option("--nbits", default=200, show_default=True, help="Number of bits to run in simulations.")
@click.option("--plot-t-max", default=0.5e-9, show_default=True, help="Maximum time value for plots (s).")
@click.option("--f-max",  default=40e9, show_default=True, help="Maximum frequency for transfer functions (Hz).")
@click.option("--f-step", default=10e6, show_default=True, help="Frequency step for transfer functions (Hz).")
@click.option("--fig-x", default=10, show_default=True, help="x-dimmension for plot figures (in).")
@click.option("--fig-y", default=3, show_default=True, help="y-dimmension for plot figures (in).")
@click.argument("ibis_file", type=click.Path(exists=True))
@click.argument("bit_rate", type=float)
@click.version_option(package_name="PyIBIS-AMI")
# pylint: disable=too-many-arguments,too-many-positional-arguments
def main(notebook, out_dir, params, ibis_file, bit_rate,  # pylint: disable=too-many-locals
         debug, is_tx, nspui, no_nspui_swp, nbits,
         plot_t_max, f_max, f_step, fig_x, fig_y):
    "Run a *Jupyter* notebook on an IBIS-AMI model file."
    arguments_list = sys.argv
    full_command_line = "run-notebook " + " ".join(shlex.quote(arg) for arg in arguments_list[1:])
    run_notebook(
        Path(ibis_file).resolve(), Path(notebook).resolve(),
        out_dir=Path(out_dir).resolve(),
        notebook_params={
            'ibis_dir': ".",
            'ibis_file': ibis_file,
            'debug': debug,
            'is_tx': is_tx,
            'nspui': nspui,
            'no_nspui_swp': no_nspui_swp,
            'nbits': nbits,
            'plot_t_max': plot_t_max,
            'f_max': f_max,
            'f_step': f_step,
            'fig_x': fig_x,
            'fig_y': fig_y,
            'bit_rate': bit_rate,
            'params': params,
            'full_command_line': full_command_line,
        })


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""A package of Python modules, used to configure and test IBIS-AMI models.

.. moduleauthor:: David Banas <capn.freako@gmail.com>

Original Author: David Banas <capn.freako@gmail.com>

Original Date:   3 July 2012

Copyright (c) 2012 by David Banas; All rights reserved World wide.
"""

from importlib.metadata import version as _get_version

# Set PEP396 version attribute
try:
    __version__ = _get_version("PyIBIS-AMI")
except Exception as err:  # pylint: disable=broad-exception-caught
    __version__ = f"{err} (dev)"

__date__ = "October 12, 2023"
__authors__ = "David Banas & David Patterson"
__copy__ = "Copyright (c) 2012 David Banas, 2019 David Patterson"



"""Parse an IBIS model file.

Original Author: David Banas <capn.freako@gmail.com>

Original Date:   November 1, 2019

For information regarding the IBIS modeling standard, visit:
https://ibis.org/

Copyright (c) 2019 by David Banas; All rights reserved World wide.
"""

import re

from parsec import (
    ParseError,
    Parser,
    count,
    eof,
    fail_with,
    generate,
    letter,
    many,
    many1,
    none_of,
    one_of,
    optional,
    regex,
    separated,
    sepBy1,
    string,
    times,
)

from pyibisami.ibis.model import Component, Model

DEBUG = False

# Parser Definitions

whitespace = regex(r"\s+", re.MULTILINE)
comment    = regex(r"\|.*")                 # To end of line only.
ignore     = many(whitespace | comment)     # None is okay; so, can be used completely safely.


def logf(p: Parser, preStr: str = "") -> Parser:
    """
    Returns parser ``p`` wrapped in a thin shell, which logs any failure at the point of occurence.

    Args:
        p: The original parser.

    Keyword Args:
        preStr: A prefix string to use in failure message.
            (Default = <empty string>)

    Returns:
        p': The original parser wrapped in a thin failure location shell.
    """

    @Parser
    def fn(txt, ix):
        res = p(txt, ix)
        if not res.status:
            print(
                f"{preStr}: Expected `{res.expected}` in `{txt[res.index: res.index + 5]}` at {ParseError.loc_info(txt, res.index)}.",
                flush=True
            )
        return res

    return fn


def lexeme(p):
    """Lexer for words.

    Skips all ignored characters after word, including newlines.
    """
    return p << ignore


def word(p):
    """Line limited word lexer.

    Only skips space after words; dosen't skip comments or newlines.
    Requires, at least, one white space character after word.
    """
    return p << regex(r"\s+")


@generate("remainder of line")
def rest_line():
    "Parse remainder of line."
    chars = yield many(none_of("[\n\r")) << ignore  # So that we still function as a lexeme.
    rslt = "".join(chars)
    return rslt


skip_line = lexeme(rest_line).result("(Skipped.)")
name_only = regex(r"[_a-zA-Z0-9/\.()#-]+")
name = word(name_only)
symbol = lexeme(regex(r"[a-zA-Z_][^\s()\[\]]*"))
true = lexeme(string("True")).result(True)
false = lexeme(string("False")).result(False)
quoted_string = lexeme(regex(r'"[^"]*"'))
skip_keyword = (skip_line >> many(none_of("[") >> skip_line)).result(
    "(Skipped.)"
)  # Skip over everything until the next keyword begins.

IBIS_num_suf = {
    "T": "e12",
    "k": "e3",
    "n": "e-9",
    "G": "e9",
    "m": "e-3",
    "p": "e-12",
    "M": "e6",
    "u": "e-6",
    "f": "e-15",
}


@generate("number")
def number():
    "Parse an IBIS numerical value."
    s = yield (regex(r"[-+]?[0-9]*\.?[0-9]+(([eE][-+]?[0-9]+)|([TknGmpMuf][a-zA-Z]*))?") << many(letter()) << ignore)
    m = re.search(r"[^\d]+$", s)
    if m:
        ix = m.start()
        c = s[ix]
        if c in IBIS_num_suf:
            res = float(s[:ix] + IBIS_num_suf[c])
        else:
            raise ParseError("IBIS numerical suffix", s[ix:], ix)
    else:
        res = float(s)
    return res


na = word(string("NA") | string("na")).result(None)


@generate("typminmax")
def typminmax():
    "Parse Typ/Min/Max values."
    typ = yield number
    minmax = yield optional(count(number, 2) | count(na, 2).result([]), [])
    yield ignore  # So that ``typminmax`` behaves as a lexeme.
    res = [typ]
    res.extend(minmax)
    return res


vi_line = (number + typminmax) << ignore


@generate("ratio")
def ratio():
    "Parse ratio."
    [num, den] = yield (separated(number, string("/"), 2, maxt=2, end=False) | na.result([0, 0]))
    if den:
        return num / den
    return None


ex_line = (
    word(string("Executable"))
    >> (  # noqa: W503
        (
            ((string("L") | string("l")) >> string("inux")).result("linux")
            | ((string("W") | string("w")) >> string("indows")).result("windows")  # noqa: W503
        )
        << string("_")  # noqa: W503
        << many(none_of("_"))  # noqa: W503
        << string("_")  # noqa: W503
    )
    + lexeme(string("32") | string("64"))  # noqa: W503
    + count(name, 2)  # noqa: W503
    << ignore  # noqa: W503
)


def manyTrue(p):
    "Run a parser multiple times, filtering ``False`` results."

    @generate("manyTrue")
    def fn():
        "many(p) >> filter(True)"
        nodes = yield many(p)
        res = list(filter(None, nodes))
        return res

    return fn


def many1True(p):
    "Run a parser at least once, filtering ``False`` results."

    @generate("many1True")
    def fn():
        "many1(p) >> filter(True)"
        nodes = yield many1(p)
        res = list(filter(None, nodes))
        return res

    return fn


# IBIS file parser:


def keyword(kywrd=""):
    """Parse an IBIS keyword.

    Keyword Args:
        kywrd (str): The particular keyword to match; null for any keyword.
            If provided, *must* be in canonicalized form (i.e. - underscores,
            no spaces)!

    Returns:
        Parser: A keyword parser.
    """

    @generate("IBIS keyword")
    def fn():
        "Parse IBIS keyword."
        yield regex(r"^\[", re.MULTILINE)
        wordlets = yield sepBy1(name_only, one_of(" _"))  # ``name`` gobbles up trailing space, which we don't want.
        yield string("]")
        yield ignore  # So that ``keyword`` functions as a lexeme.
        res = "_".join(wordlets)  # Canonicalize to: "<wordlet1>_<wordlet2>_...".
        if kywrd:
            if res.lower() == kywrd.lower():
                return res
            return fail_with(f"Expecting: {kywrd}; got: {res}.")
        return res

    return fn


@generate("IBIS parameter")
def param():
    "Parse IBIS parameter."
    # Parameters must begin with a letter in column 1.
    pname = yield word(regex(r"^[a-zA-Z]\w*", re.MULTILINE))
    if DEBUG:
        print(f"Parsing parameter {pname}...", end="", flush=True)
    res = yield ((word(string("=")) >> (number | rest_line)) | typminmax | name | rest_line)
    if DEBUG:
        print(res, flush=True)
    yield ignore  # So that ``param`` functions as a lexeme.
    return (pname.lower(), res)


def node(valid_keywords, stop_keywords, debug=False):
    """Build a node-specific parser.

    Args:
        valid_keywords (dict): A dictionary with keys matching those
            keywords we want parsed. The values are the parsers for
            those keywords.
        stop_keywords: Any iterable with primary values (i.e. - those
            tested by the ``in`` function) matching those keywords we want
            to stop the parsing of this node and pop us back up the
            parsing stack.

    Returns:
        Parser: A parser for this node.

    Notes:
        1: Any keywords encountered that are _not_ found (via ``in``) in
            either ``valid_keywords`` or ``stop_keywords`` are ignored.
    """

    @generate("kywrd")
    def kywrd():
        "Parse keyword syntax."
        nm = yield keyword()
        nmL = nm.lower()
        if debug:
            print(f"Parsing keyword: [{nm}]...", flush=True)
        if nmL in valid_keywords:
            if nmL == "end":  # Because ``ibis_file`` expects this to be the last thing it sees,
                return fail_with("")  # we can't consume it here.
            res = yield logf(valid_keywords[nmL], f"[{nm}]")  # Parse the sub-keyword.
        elif nmL in stop_keywords:
            return fail_with("")  # Stop parsing.
        else:
            res = yield skip_keyword
        yield ignore  # So that ``kywrd`` behaves as a lexeme.
        if debug:
            print(f"Finished parsing keyword: [{nm}].", flush=True)
        return (nmL, res)

    return kywrd | param


# Individual IBIS keyword (i.e. - "node") parsers:


# [End]
@generate("[End]")
def end():
    "Parse [End]."
    yield keyword("End")
    return eof


# [Model]
load_line = string("R_load = ") >> number

ramp_line = string("dV/dt_") >> ((string("r").result("rising") | string("f").result("falling")) << ignore) + times(
    ratio, 1, 3
)


@generate("[Ramp]")
def ramp():
    "Parse [Ramp]."
    load = yield optional(load_line, 50)
    lines = yield count(ramp_line, 2).desc("Two ramp_lines")
    rslt = dict(lines)
    rslt.update({"load": load})
    return rslt


Model_keywords = {
    "pulldown": many1(vi_line),
    "pullup": many1(vi_line),
    "ramp": ramp,
    "algorithmic_model": many1(ex_line) << keyword("end_algorithmic_model"),
    "voltage_range": typminmax,
    "temperature_range": typminmax,
    "gnd_clamp": many1(vi_line),
    "power_clamp": many1(vi_line),
}


@generate("[Model]")
def model():
    "Parse [Model]."
    nm = yield name << ignore
    if DEBUG:
        print(f"Parsing model: {nm}...", flush=True)
    res = yield many1(node(Model_keywords, IBIS_keywords, debug=DEBUG))
    if DEBUG:
        print(f"[Model] {nm} contains: {dict(res).keys()}", flush=True)
    try:
        theModel = Model(dict(res))
    except LookupError as le:
        return fail_with(f"[Model] {nm}: {str(le)}")
    return {nm: theModel}


# [Component]
rlc = lexeme(string("R_pin") | string("L_pin") | string("C_pin"))


@generate("[Manufacturer]")
def manufacturer():
    "Parse manufacturer."
    rslt = yield rest_line
    if not rslt:
        rslt = "(n/a)"
    if DEBUG:
        print(f"Manufacturer: {rslt}", flush=True)
    return rslt


@generate("[Package]")
def package():
    "Parse package RLC values."
    rlcs = yield many1(param)
    if DEBUG:
        print(f"rlcs: {rlcs}", flush=True)
    return dict(rlcs)


def pin(rlcs):
    "Parse indiviual component pin."

    @generate("Component Pin")
    def fn():
        "Parse an individual component pin."
        [nm, sig] = yield count(name, 2)
        mod = yield name_only
        rem_line = yield rest_line
        rlc_vals = optional(count(number, 3), []).parse(rem_line)
        rlc_dict = {}
        if rlcs:
            rlc_dict.update(dict(zip(rlcs, rlc_vals)))
        return ((nm + "(" + sig + ")"), (mod, rlc_dict))

    return fn


@generate("[Component].[Pin]")
def pins():
    "Parse [Component].[Pin]."

    def filt(x):
        (_, (mod, _)) = x
        m = mod.upper()
        return m not in ("POWER", "GND", "NC")

    yield (lexeme(string("signal_name")) << lexeme(string("model_name")))
    rlcs = yield optional(count(rlc, 3), [])
    prs = yield many1(pin(rlcs))
    prs_filt = list(filter(filt, prs))
    return dict(prs_filt)


Component_keywords = {
    "manufacturer": manufacturer,
    "package": package,
    "pin": pins,
    "diff_pin": skip_keyword,
}


@generate("[Component]")
def comp():
    "Parse [Component]."
    nm = yield lexeme(name)
    if DEBUG:
        print(f"Parsing component: {nm}", flush=True)
    res = yield many1(node(Component_keywords, IBIS_keywords, debug=DEBUG))
    try:
        Component(dict(res))
    except LookupError as le:
        return fail_with(f"[Component] {nm}: {str(le)}")
    except Exception as err:  # pylint: disable=broad-exception-caught
        return fail_with(f"[Component] {nm}: {str(err)}")
    return {nm: Component(dict(res))}


@generate("[Model Selector]")
def modsel():
    "Parse [Model Selector]."
    nm = yield name
    res = yield ignore >> many1(name + rest_line)
    return {nm: res}


# Note: The following list MUST have a complete set of keys,
#       in order for the parsing logic to work correctly!
IBIS_keywords = [
    "model",
    "end",
    "ibis_ver",
    "comment_char",
    "file_name",
    "file_rev",
    "date",
    "source",
    "notes",
    "disclaimer",
    "copyright",
    "component",
    "model_selector",
    "submodel",
    "external_circuit",
    "test_data",
    "test_load",
    "define_package_model",
    "interconnect_model_set",
]

IBIS_kywrd_parsers = dict(zip(IBIS_keywords, [skip_keyword] * len(IBIS_keywords)))
IBIS_kywrd_parsers.update(
    {
        "model": model,
        "end": end,
        "ibis_ver": lexeme(number),
        "file_name": lexeme(name),
        "file_rev": lexeme(name),
        "date": rest_line,
        "component": comp,
        "model_selector": modsel,
    }
)


@generate("IBIS File")
def ibis_file():
    "Parse IBIS file."
    res = yield ignore >> many1True(node(IBIS_kywrd_parsers, {}, debug=DEBUG)) << end
    return res


def parse_ibis_file(ibis_file_contents_str, debug=False):
    """Parse the contents of an IBIS file.

    Args:
        ibis_file_contents_str (str): The contents of the IBIS file, as a single string.

    Keyword Args:
        debug (bool): Output debugging info to console when true.
            Default = False

    Example:
        ::

            with open(<ibis_file_name>) as ibis_file:
                ibis_file_contents_str = ibis_file.read()
                (err_str, model_dict)  = parse_ibis_file(ibis_file_contents_str)

    Returns:
        (str, dict): A pair containing:

            err_str:
                A message describing the nature of any parse failure that occured.
            model_dict:
                Dictionary containing keyword definitions (empty upon failure).
    """

    global DEBUG  # pylint: disable=W0603
    DEBUG = debug

    try:
        nodes = ibis_file.parse_strict(ibis_file_contents_str)  # Parse must consume the entire file.
        if debug:
            print("Parsed nodes:\n", nodes, flush=True)
    except ParseError as pe:
        return str(pe), {}

    kw_dict = {}
    components = {}
    models = {}
    model_selectors = {}
    for kw, val in nodes:
        if kw == "model":
            models.update(val)
        elif kw == "component":
            components.update(val)
        elif kw == "model_selector":
            model_selectors.update(val)
        else:
            kw_dict.update({kw: val})
    kw_dict.update(
        {
            "components": components,
            "models": models,
            "model_selectors": model_selectors,
        }
    )
    return "Success!", kw_dict



"""Classes for encapsulating IBIS model constituents.

Original Author: David Banas <capn.freako@gmail.com>

Original Date:   November 1, 2019

For information regarding the IBIS modeling standard, visit:
https://ibis.org/

Copyright (c) 2019 by David Banas; All rights reserved World wide.
"""

import numpy as np
from chaco.api import ArrayPlotData, Plot
from enable.component_editor import ComponentEditor
from traits.api import HasTraits, String, Trait
from traitsui.api import Group, Item, ModalButtons, View


class Component(HasTraits):
    """Encapsulation of a particular component from an IBIS model file."""

    def __init__(self, subDict):
        """
        Args:
            subDict(dict): Dictionary of [Component] sub-keywords/params.
        """

        # Super-class initialization is ABSOLUTELY NECESSARY, in order
        # to get all the Traits/UI machinery setup correctly.
        super().__init__()

        # Stash the sub-keywords/parameters.
        self._subDict = subDict

        # Fetch available keyword/parameter definitions.
        def maybe(name):
            return subDict[name] if name in subDict else None

        self._mfr = maybe("manufacturer")
        self._pkg = maybe("package")
        self._pins = maybe("pin")
        self._diffs = maybe("diff_pin")

        # Check for the required keywords.
        if not self._mfr:
            raise LookupError("Missing [Manufacturer]!")
        if not self._pkg:
            print(self._mfr)
            raise LookupError("Missing [Package]!")
        if not self._pins:
            raise LookupError("Missing [Pin]!")

        # Set up the GUI.
        self.add_trait("manufacturer", String(self._mfr))
        self.add_trait("package", String(self._pkg))
        self.add_trait("_pin", Trait(list(self._pins)[0], self._pins))
        self._content = [
            Group(
                Item("manufacturer", label="Manufacturer", style="readonly"),
                Item("package", label="Package", style="readonly"),
                Item("_pin", label="Pin"),
                label="Component",
                show_border=True,
            ),
        ]

    def __str__(self):
        res = "Manufacturer:\t" + self._mfr + "\n"
        res += "Package:     \t" + str(self._pkg) + "\n"
        res += "Pins:\n"
        for pname in self._pins:
            res += "    " + pname + ":\t" + str(self._pins[pname]) + "\n"
        return res

    def __call__(self):
        self.edit_traits()

    def default_traits_view(self):
        "Default Traits/UI view definition."
        view = View(
            resizable=False,
            buttons=ModalButtons,
            title="PyBERT IBIS Component Viewer",
            id="pyibisami.ibis_parser.Component",
        )
        view.set_content(self._content)
        return view

    @property
    def pin(self):
        """The pin selected most recently by the user.

        Returns the first pin in the list, if the user hasn't made a
        selection yet.
        """
        return self._pin_

    @property
    def pins(self):
        "The list of component pins."
        return self._pins


class Model(HasTraits):  # pylint: disable=too-many-instance-attributes
    """Encapsulation of a particular I/O model from an IBIS model file."""

    def __init__(self, subDict):  # pylint: disable=too-many-locals,too-many-statements,too-many-branches
        """
        Args:
            subDict (dict): Dictionary of sub-keywords/params.
        """

        # Super-class initialization is ABSOLUTELY NECESSARY, in order
        # to get all the Traits/UI machinery setup correctly.
        super().__init__()

        # Stash the sub-keywords/parameters.
        self._subDict = subDict

        # Fetch available keyword/parameter definitions.
        def maybe(name):
            return subDict[name] if name in subDict else None

        self._mtype = maybe("model_type")
        self._ccomp = maybe("c_comp")
        self._cref = maybe("cref")
        self._vref = maybe("vref")
        self._vmeas = maybe("vmeas")
        self._rref = maybe("rref")
        self._trange = maybe("temperature_range")
        self._vrange = maybe("voltage_range")
        self._ramp = maybe("ramp")

        # Check for the required keywords.
        if not self._mtype:
            raise LookupError("Missing Model_type!")
        if not self._vrange:
            raise LookupError("Missing [Voltage Range]!")

        def proc_iv(xs):
            """Process an I/V table."""
            if len(xs) < 2:
                raise ValueError("Insufficient number of I-V data points!")
            try:
                vs, iss = zip(*(xs))  # Idiomatic Python for ``unzip``.
            except Exception as exc:
                raise ValueError(f"xs: {xs}") from exc
            ityps, imins, imaxs = zip(*iss)
            vmeas = self._vmeas

            def calcZ(x):
                (vs, ivals) = x
                if vmeas:
                    ix = np.where(np.array(vs) >= vmeas)[0][0]
                else:
                    ix = np.where(np.array(vs) >= max(vs) / 2)[0][0]
                try:
                    return abs((vs[ix] - vs[ix - 1]) / (ivals[ix] - ivals[ix - 1]))
                except ZeroDivisionError:
                    return 1e7  # Use 10 MOhms in place of infinity.

            zs = map(calcZ, zip([vs, vs, vs], [ityps, imins, imaxs]))
            return vs, ityps, imins, imaxs, zs

        # Infer impedance and/or rise/fall time, as per model type.
        mtype = self._mtype.lower()
        if mtype in ("output", "i/o"):
            if "pulldown" not in subDict or "pullup" not in subDict:
                raise LookupError("Missing I-V curves!")
            plotdata = ArrayPlotData()
            pd_vs, pd_ityps, pd_imins, pd_imaxs, pd_zs = proc_iv(subDict["pulldown"])
            pu_vs, pu_ityps, pu_imins, pu_imaxs, pu_zs = proc_iv(subDict["pullup"])
            pu_vs = self._vrange[0] - np.array(pu_vs)  # Correct for Vdd-relative pull-up voltages.
            pu_ityps = -np.array(pu_ityps)  # Correct for current sense, for nicer plot.
            pu_imins = -np.array(pu_imins)
            pu_imaxs = -np.array(pu_imaxs)
            self._zout = (list(pd_zs)[0] + list(pu_zs)[0]) / 2
            plotdata.set_data("pd_vs", pd_vs)
            plotdata.set_data("pd_ityps", pd_ityps)
            plotdata.set_data("pd_imins", pd_imins)
            plotdata.set_data("pd_imaxs", pd_imaxs)
            plotdata.set_data("pu_vs", pu_vs)
            plotdata.set_data("pu_ityps", pu_ityps)
            plotdata.set_data("pu_imins", pu_imins)
            plotdata.set_data("pu_imaxs", pu_imaxs)
            plot_iv = Plot(plotdata)  # , padding_left=75)
            # The 'line_style' trait of a LinePlot instance must be 'dash' or 'dot dash' or 'dot' or 'long dash' or 'solid'.
            plot_iv.plot(("pd_vs", "pd_ityps"), type="line", color="blue", line_style="solid", name="PD-Typ")
            plot_iv.plot(("pd_vs", "pd_imins"), type="line", color="blue", line_style="dot", name="PD-Min")
            plot_iv.plot(("pd_vs", "pd_imaxs"), type="line", color="blue", line_style="dash", name="PD-Max")
            plot_iv.plot(("pu_vs", "pu_ityps"), type="line", color="red", line_style="solid", name="PU-Typ")
            plot_iv.plot(("pu_vs", "pu_imins"), type="line", color="red", line_style="dot", name="PU-Min")
            plot_iv.plot(("pu_vs", "pu_imaxs"), type="line", color="red", line_style="dash", name="PU-Max")
            plot_iv.title = "Pull-Up/Down I-V Curves"
            plot_iv.index_axis.title = "Vout (V)"
            plot_iv.value_axis.title = "Iout (A)"
            plot_iv.index_range.low_setting = 0
            plot_iv.index_range.high_setting = self._vrange[0]
            plot_iv.value_range.low_setting = 0
            plot_iv.value_range.high_setting = 0.1
            plot_iv.legend.visible = True
            plot_iv.legend.align = "ul"
            self.plot_iv = plot_iv

            if not self._ramp:
                raise LookupError("Missing [Ramp]!")
            ramp = subDict["ramp"]
            self._slew = (ramp["rising"][0] + ramp["falling"][0]) / 2e9  # (V/ns)
        elif mtype == "input":
            if "gnd_clamp" not in subDict and "power_clamp" not in subDict:
                # raise LookupError("Missing clamp curves!")
                pass

            plotdata = ArrayPlotData()

            if "gnd_clamp" in subDict:
                gc_vs, gc_ityps, gc_imins, gc_imaxs, gc_zs = proc_iv(subDict["gnd_clamp"])
                gc_z = list(gc_zs)[0]  # Use typical value for Zin calc.
                plotdata.set_data("gc_vs", gc_vs)
                plotdata.set_data("gc_ityps", gc_ityps)
                plotdata.set_data("gc_imins", gc_imins)
                plotdata.set_data("gc_imaxs", gc_imaxs)
            else:
                gc_z = 2e6

            if "power_clamp" in subDict:
                pc_vs, pc_ityps, pc_imins, pc_imaxs, pc_zs = proc_iv(subDict["power_clamp"])
                pc_z = list(pc_zs)[0]
                pc_vs = self._vrange[0] - np.array(pc_vs)  # Correct for Vdd-relative pull-up voltages.
                pc_ityps = -np.array(pc_ityps)  # Correct for current sense, for nicer plot.
                pc_imins = -np.array(pc_imins)
                pc_imaxs = -np.array(pc_imaxs)
                plotdata.set_data("pc_vs", pc_vs)
                plotdata.set_data("pc_ityps", pc_ityps)
                plotdata.set_data("pc_imins", pc_imins)
                plotdata.set_data("pc_imaxs", pc_imaxs)
            else:
                pc_z = 2e6

            plot_iv = Plot(plotdata)  # , padding_left=75)
            # The 'line_style' trait of a LinePlot instance must be 'dash' or 'dot dash' or 'dot' or 'long dash' or 'solid'.
            if "gnd_clamp" in subDict:
                plot_iv.plot(("gc_vs", "gc_ityps"), type="line", color="blue", line_style="solid", name="PD-Typ")
                plot_iv.plot(("gc_vs", "gc_imins"), type="line", color="blue", line_style="dot", name="PD-Min")
                plot_iv.plot(("gc_vs", "gc_imaxs"), type="line", color="blue", line_style="dash", name="PD-Max")
            if "power_clamp" in subDict:
                plot_iv.plot(("pc_vs", "pc_ityps"), type="line", color="red", line_style="solid", name="PU-Typ")
                plot_iv.plot(("pc_vs", "pc_imins"), type="line", color="red", line_style="dot", name="PU-Min")
                plot_iv.plot(("pc_vs", "pc_imaxs"), type="line", color="red", line_style="dash", name="PU-Max")
            plot_iv.title = "Power/GND Clamp I-V Curves"
            plot_iv.index_axis.title = "Vin (V)"
            plot_iv.value_axis.title = "Iin (A)"
            plot_iv.index_range.low_setting = 0
            plot_iv.index_range.high_setting = self._vrange[0]
            plot_iv.value_range.low_setting = 0
            plot_iv.value_range.high_setting = 0.1
            plot_iv.legend.visible = True
            plot_iv.legend.align = "ul"
            self.plot_iv = plot_iv

            if "gnd_clamp" in subDict and "power_clamp" in subDict:
                # Parallel combination, as both clamps are always active.
                self._zin = (gc_z * pc_z) / (gc_z + pc_z)  # pylint: disable=possibly-used-before-assignment
            elif "gnd_clamp" in subDict:
                self._zin = gc_z
            else:
                self._zin = pc_z

        # Separate AMI executables by OS.
        def is64(x):
            ((_, b), _) = x
            return int(b) == 64

        def isWin(x):
            ((os, _), _) = x
            return os.lower() == "windows"

        def partition(p, xs):
            ts, fs = [], []
            for x in xs:
                if p(x):
                    ts.append(x)
                else:
                    fs.append(x)
            return ts, fs

        def getFiles(x):
            if x:
                ((_, _), fs) = x[0]
                return fs
            return []

        def splitExecs(fs):
            wins, lins = partition(isWin, fs)
            return (getFiles(wins), getFiles(lins))

        self._exec32Wins, self._exec32Lins = [], []
        self._exec64Wins, self._exec64Lins = [], []
        if "algorithmic_model" in subDict:
            execs = subDict["algorithmic_model"]
            exec64s, exec32s = partition(is64, execs)
            self._exec32Wins, self._exec32Lins = splitExecs(exec32s)
            self._exec64Wins, self._exec64Lins = splitExecs(exec64s)

        # Set up the GUI.
        self.add_trait("model_type", String(self._mtype))
        self.add_trait("c_comp", String(self._ccomp))
        self.add_trait("cref", String(self._cref))
        self.add_trait("vref", String(self._vref))
        self.add_trait("vmeas", String(self._vmeas))
        self.add_trait("rref", String(self._rref))
        self.add_trait("trange", String(self._trange))
        self.add_trait("vrange", String(self._vrange))
        if mtype in ("output", "i/o"):
            self.add_trait("zout", String(self._zout))
            self.add_trait("slew", String(self._slew))
        elif mtype == "input":
            self.add_trait("zin", String(self._zin))
        self._content = [
            Group(
                Item("model_type", label="Model type", style="readonly"),
                Item("c_comp", label="Ccomp", style="readonly"),
                Item("trange", label="Temperature Range", style="readonly"),
                Item("vrange", label="Voltage Range", style="readonly"),
                Group(
                    Item("cref", label="Cref", style="readonly"),
                    Item("vref", label="Vref", style="readonly"),
                    Item("vmeas", label="Vmeas", style="readonly"),
                    Item("rref", label="Rref", style="readonly"),
                    orientation="horizontal",
                ),
                label="Model",
                show_border=True,
            ),
        ]
        if mtype in ("output", "i/o"):
            self._content.append(Item("zout", label="Impedance (Ohms)", style="readonly", format_str="%4.1f"))
            self._content.append(Item("slew", label="Slew Rate (V/ns)", style="readonly", format_str="%4.1f"))
            self._content.append(Item("plot_iv", editor=ComponentEditor(), show_label=False))
        elif mtype == "input":
            self._content.append(Item("zin", label="Impedance (Ohms)", style="readonly", format_str="%4.1f"))
            self._content.append(Item("plot_iv", editor=ComponentEditor(), show_label=False))

    def __str__(self):
        res = "Model Type:\t" + self._mtype + "\n"
        res += "C_comp:    \t" + str(self._ccomp) + "\n"
        res += "Cref:      \t" + str(self._cref) + "\n"
        res += "Vref:      \t" + str(self._vref) + "\n"
        res += "Vmeas:     \t" + str(self._vmeas) + "\n"
        res += "Rref:      \t" + str(self._rref) + "\n"
        res += "Temperature Range:\t" + str(self._trange) + "\n"
        res += "Voltage Range:    \t" + str(self._vrange) + "\n"
        if "algorithmic_model" in self._subDict:
            res += "Algorithmic Model:\n" + "\t32-bit:\n"
            if self._exec32Lins:
                res += "\t\tLinux: " + str(self._exec32Lins) + "\n"
            if self._exec32Wins:
                res += "\t\tWindows: " + str(self._exec32Wins) + "\n"
            res += "\t64-bit:\n"
            if self._exec64Lins:
                res += "\t\tLinux: " + str(self._exec64Lins) + "\n"
            if self._exec64Wins:
                res += "\t\tWindows: " + str(self._exec64Wins) + "\n"
        return res

    def __call__(self):
        self.edit_traits(kind="livemodal")

    def default_traits_view(self):
        "Default Traits/UI view definition."
        view = View(
            resizable=False,
            buttons=ModalButtons,
            title="PyBERT IBIS Model Viewer",
            id="pyibisami.ibis_parser.Model",
        )
        view.set_content(self._content)
        return view

    @property
    def zout(self):
        "The driver impedance."
        return self._zout

    @property
    def slew(self):
        "The driver slew rate."
        return self._slew

    @property
    def zin(self):
        "The input impedance."
        return self._zin

    @property
    def ccomp(self):
        "The parasitic capacitance."
        return self._ccomp

    @property
    def mtype(self):
        """Model type."""
        return self._mtype



"""A class for encapsulating IBIS model files.

Original Author: David Banas <capn.freako@gmail.com>

Original Date:   November 1, 2019

For information regarding the IBIS modeling standard, visit:
https://ibis.org/

**Note:** The ``IBISModel`` class, defined here, needs to be kept separate from the
other IBIS-related classes, defined in the ``model`` module, in order to
avoid circular imports.

Copyright (c) 2019 by David Banas; All rights reserved World wide.
"""

import platform
from datetime import datetime

from traits.api import (
    Any,
    Dict,
    Enum,
    Float,
    HasTraits,
    List,
    Property,
    String,
    Trait,
    cached_property,
)
from traitsui.api import HGroup, Item, ModalButtons, VGroup, View, spring
from traitsui.message import message

from pyibisami.ibis.parser import parse_ibis_file


class IBISModel(HasTraits):  # pylint: disable=too-many-instance-attributes
    """HasTraits subclass for wrapping and interacting with an IBIS model.

    This class can be configured to present a customized GUI to the user
    for interacting with a particular IBIS model (i.e. - selecting components,
    pins, and models).

    The intended use model is as follows:

     1. Instantiate this class only once per IBIS model file.
        When instantiating, provide the unprocessed contents of the IBIS
        file, as a single string. This class will take care of getting
        that string parsed properly, and report any errors or warnings
        it encounters, in its ``ibis_parsing_errors`` property.

     2. When you want to let the user select a particular component/pin/model,
        call the newly created instance, as if it were a function, passing
        no arguments.
        The instance will then present a GUI to the user,
        allowing him to select a particular component/pin/model, which may then
        be retrieved, via the ``model`` property.
        The latest user selections will be remembered,
        as long as the instance remains in scope.

    Any errors or warnings encountered while parsing are available, in
    the ``ibis_parsing_errors`` property.

    The complete dictionary containing all parsed models may be retrieved,
    via the ``model_dict`` property.
    """

    _log = ""

    pin_ = Property(Any, depends_on=["pin"])
    pin_rlcs = Property(Dict, depends_on=["pin"])
    model = Property(Any, depends_on=["mod"])
    pins = List  # Always holds the list of valid pin selections, given a component selection.
    models = List  # Always holds the list of valid model selections, given a pin selection.

    def get_models(self, mname):
        """Return the list of models associated with a particular name."""
        model_dict = self._model_dict
        if "model_selectors" in model_dict and mname in model_dict["model_selectors"]:
            return list(map(lambda pr: pr[0], model_dict["model_selectors"][mname]))
        return [mname]

    def get_pins(self):
        """Get the list of appropriate pins, given our type (i.e. - Tx or Rx)."""
        pins = self.comp_.pins

        def pin_ok(pname):
            (mname, _) = pins[pname]
            mods = self.get_models(mname)
            mod = self._models[mods[0]]
            mod_type = mod.mtype.lower()
            tx_ok = mod_type in ("output", "i/o")
            if self._is_tx:
                return tx_ok
            return not tx_ok

        return list(filter(pin_ok, list(pins)))

    def __init__(self, ibis_file_name, is_tx, debug=False, gui=True):
        """
        Args:
            ibis_file_name (str): The name of the IBIS file.
            is_tx (bool): True if this is a Tx model.

        Keyword Args:
            debug (bool): Output debugging info to console when true.
                Default = False
            gui (bool): Set to `False` for command line and/or script usage.
                Default = True.
        """

        # Super-class initialization is ABSOLUTELY NECESSARY, in order
        # to get all the Traits/UI machinery setup correctly.
        super().__init__()

        self.debug = debug
        self.GUI = gui
        if debug:
            self.log("pyibisami.ibis_file.IBISModel initializing in debug mode...")
        else:
            self.log("pyibisami.ibis_file.IBISModel initializing in non-debug mode...")

        # Parse the IBIS file contents, storing any errors or warnings, and validate it.
        with open(ibis_file_name, "r", encoding="utf-8") as file:
            ibis_file_contents_str = file.read()
        err_str, model_dict = parse_ibis_file(ibis_file_contents_str, debug=debug)
        self.log("IBIS parsing errors/warnings:\n" + err_str)
        if "components" not in model_dict or not model_dict["components"]:
            print(f":\n{model_dict}", flush=True)
            raise ValueError("This IBIS model has no components!")
        components = model_dict["components"]
        if "models" not in model_dict or not model_dict["models"]:
            raise ValueError("This IBIS model has no models!")
        models = model_dict["models"]
        self._model_dict = model_dict
        self._models = models
        self._is_tx = is_tx

        # Add Traits for various attributes found in the IBIS file.
        self.add_trait("comp", Trait(list(components)[0], components))  # Doesn't need a custom mapper, because
        self.pins = self.get_pins()                                     # the thing above it (file) can't change.
        self.add_trait("pin", Enum(self.pins[0], values="pins"))
        (mname, _) = self.pin_
        self.models = self.get_models(mname)
        self.add_trait("mod", Enum(self.models[0], values="models"))
        self.add_trait("ibis_ver", Float(model_dict["ibis_ver"]))
        self.add_trait("file_name", String(model_dict["file_name"]))
        self.add_trait("file_rev", String(model_dict["file_rev"]))
        if "date" in model_dict:
            self.add_trait("date", String(model_dict["date"]))
        else:
            self.add_trait("date", String("(n/a)"))

        self._ibis_parsing_errors = err_str
        self._os_type = platform.system()  # These 2 are used, to choose
        self._os_bits = platform.architecture()[0]  # the correct AMI executable.

        self._comp_changed(list(components)[0])  # Wasn't being called automatically.
        self._pin_changed(self.pins[0])  # Wasn't being called automatically.

        self.log("Done.")

    def __str__(self):
        return f"IBIS Model '{self._model_dict['file_name']}'"

    def info(self):
        """Basic information about the IBIS model."""
        res = ""
        try:
            for k in ["ibis_ver", "file_name", "file_rev"]:
                res += k + ":\t" + str(self._model_dict[k]) + "\n"
        except Exception as err:
            print(f"{err}")
            print(self._model_dict)
            raise
        res += "date" + ":\t\t" + str(self._model_dict["date"]) + "\n"
        res += "\nComponents:"
        res += "\n=========="
        for c in list(self._model_dict["components"]):
            res += "\n" + c + ":\n" + "---\n" + str(self._model_dict["components"][c]) + "\n"
        res += "\nModel Selectors:"
        res += "\n===============\n"
        for s in list(self._model_dict["model_selectors"]):
            res += f"{s}\n"
        res += "\nModels:"
        res += "\n======"
        for m in list(self._model_dict["models"]):
            res += "\n" + m + ":\n" + "---\n" + str(self._model_dict["models"][m])
        return res

    def __call__(self):
        """Present a customized GUI to the user, for model selection, etc."""
        # self.configure_traits(kind='modal')  # Waiting for Enthought/Traits PR1841 to be accepted.
        self.configure_traits()

    # Logger & Pop-up
    def log(self, msg, alert=False):
        """Log a message to the console and, optionally, to terminal and/or
        pop-up dialog."""
        _msg = msg.strip()
        txt = f"\n[{datetime.now()}]: IBISModel: {_msg}\n"
        self._log += txt
        if self.debug:
            print(txt, flush=True)
        if alert and self.GUI:
            message(_msg, "PyAMI Alert")

    def default_traits_view(self):
        "Default Traits/UI view definition."
        view = View(
            VGroup(
                HGroup(
                    Item("file_name", label="File name", style="readonly"),
                    spring,
                    Item("file_rev", label="rev", style="readonly"),
                ),
                HGroup(
                    Item("ibis_ver", label="IBIS ver", style="readonly"),
                    spring,
                    Item("date", label="Date", style="readonly"),
                ),
                HGroup(
                    Item("comp", label="Component"),
                    Item("pin", label="Pin"),
                    Item("mod", label="Model"),
                ),
            ),
            resizable=False,
            buttons=ModalButtons,
            title="PyBERT IBIS Model Selector",
            id="pybert_ibis_model_selector",
        )
        return view

    @cached_property
    def _get_pin_(self):
        return self.comp_.pins[self.pin]

    @cached_property
    def _get_pin_rlcs(self):
        (_, pin_rlcs) = self.pin_
        return pin_rlcs

    @cached_property
    def _get_model(self):
        return self._models[self.mod]

    @property
    def ibis_parsing_errors(self):
        """Any errors or warnings encountered, while parsing the IBIS file
        contents."""
        return self._ibis_parsing_errors

    @property
    def log_txt(self):
        """The complete log since instantiation."""
        return self._log

    @property
    def model_dict(self):
        "Dictionary of all model keywords."
        return self._model_dict

    @property
    def dll_file(self):
        "Shared object file."
        return self._dll_file

    @property
    def ami_file(self):
        "AMI file."
        return self._ami_file

    def _comp_changed(self, new_value):
        del new_value
        self.pins = self.get_pins()
        self.pin = self.pins[0]

    def _pin_changed(self, new_value):
        # (mname, rlc_dict) = self.pin_  # Doesn't work. Because ``pin_`` is a cached property and hasn't yet been marked "dirty"?
        (mname, _) = self.comp_.pins[new_value]
        self.models = self.get_models(mname)
        self.mod = self.models[0]

    def _mod_changed(self, new_value):
        model = self._models[new_value]
        os_type = self._os_type
        os_bits = self._os_bits
        fnames = []
        dll_file = ""
        ami_file = ""
        if os_type.lower() == "windows":
            if os_bits == "64bit":
                fnames = model._exec64Wins  # pylint: disable=protected-access
            else:
                fnames = model._exec32Wins  # pylint: disable=protected-access
        else:
            if os_bits == "64bit":
                fnames = model._exec64Lins  # pylint: disable=protected-access
            else:
                fnames = model._exec32Lins  # pylint: disable=protected-access
        if fnames:
            dll_file = fnames[0]
            ami_file = fnames[1]
            self.log(
                "There was an [Algorithmic Model] keyword in this model.\n \
If you wish to use the AMI model associated with this IBIS model,\n \
please, go the 'Equalization' tab and enable it now.",
                alert=True,
            )
        elif "algorithmic_model" in model._subDict:  # pylint: disable=protected-access
            self.log(
                f"There was an [Algorithmic Model] keyword for this model,\n \
but no executable for your platform: {os_type}-{os_bits};\n \
PyBERT native equalization modeling being used instead.",
                alert=True,
            )
        else:
            self.log(
                "There was no [Algorithmic Model] keyword for this model;\n \
PyBERT native equalization modeling being used instead.",
                alert=True,
            )
        self._dll_file = dll_file  # pylint: disable=attribute-defined-outside-init
        self._ami_file = ami_file  # pylint: disable=attribute-defined-outside-init
